
import unittest
from shellish.data import hone_cache, ttl_cache, TTLMapping


class TestTTLMapping(unittest.TestCase):

    def ageless(self):
        """ Timer that never ages. """
        return 1

    def test_contains_nottl_max(self):
        c = TTLMapping(maxsize=2)
        c[1] = 11
        self.assertIn(1, c)
        c[2] = 22
        self.assertIn(2, c)
        self.assertIn(1, c)
        c[3] = 33
        self.assertIn(3, c)
        self.assertIn(2, c)
        self.assertNotIn(1, c)

    def test_contains_nottl_nomax(self):
        c = TTLMapping()
        c[1] = 11
        self.assertIn(1, c)
        c[2] = 22
        self.assertIn(2, c)
        self.assertIn(1, c)

    def test_contains_ttl_nomax(self):
        timeval = 1
        timer = lambda: timeval
        c = TTLMapping(ttl=2, timer=timer)
        c[1] = 11
        self.assertIn(1, c)
        timeval = 2
        c[2] = 22
        self.assertIn(2, c)
        self.assertIn(1, c)
        timeval = 3
        self.assertIn(2, c)
        self.assertIn(1, c)
        timeval = 4
        self.assertIn(2, c)
        self.assertNotIn(1, c)

    def test_contains_ttl_max(self):
        timeval = 1
        timer = lambda: timeval
        c = TTLMapping(ttl=2, maxsize=2, timer=timer)
        c[1] = 11
        self.assertIn(1, c)
        timeval = 2
        c[2] = 22
        self.assertIn(2, c)
        self.assertIn(1, c)
        timeval = 3
        self.assertIn(2, c)
        self.assertIn(1, c)
        timeval = 4
        self.assertIn(2, c)
        self.assertNotIn(1, c)
        c[3] = 33
        self.assertIn(2, c)
        self.assertIn(3, c)
        timeval = 5
        self.assertIn(3, c)
        self.assertNotIn(2, c)
        timeval = 7
        self.assertNotIn(3, c)
        self.assertNotIn(2, c)

    def test_size_nottl_bounding(self):
        c = TTLMapping(maxsize=1)
        self.assertEqual(len(c), 0)
        self.assertEqual(c, {})
        c[1] = 11
        self.assertEqual(len(c), 1)
        self.assertEqual(c, {1: 11})
        c[2] = 22
        self.assertEqual(len(c), 1)
        self.assertEqual(c, {2: 22})

    def test_size_ttl_bounding(self):
        for maxsize in (None, 5):
            timeval = 1
            timer = lambda: timeval
            c = TTLMapping(ttl=1, timer=timer, maxsize=maxsize)
            self.assertEqual(len(c), 0)
            self.assertEqual(c, {})
            c[1] = 1
            self.assertEqual(len(c), 1)
            self.assertEqual(c, {1: 1})
            timeval = 3
            self.assertEqual(len(c), 0)
            self.assertEqual(c, {})

    def test_iteration(self):
        for maxsize in (None, 5):
            timeval = 1
            timer = lambda: timeval
            c = TTLMapping(ttl=1, timer=timer, maxsize=128)
            c.update({1: 11, 2: 22})
            keys = list(c)
            self.assertEqual(list(keys), [1, 2])
            fast = iter(c)
            self.assertEqual(list(fast), [1, 2])
            slow = iter(c)
            timeval = 3
            self.assertEqual(list(slow), [])
            c[1] = 11
            c[2] = 22
            medium = iter(c)
            first = next(medium)
            self.assertEqual(first, 1)
            timeval = 5
            self.assertRaises(StopIteration, lambda: next(medium))


class TestTTLCache(unittest.TestCase):

    def test_pos_args(self):
        i = 0

        @ttl_cache(None)
        def test(a, b, c):
            nonlocal i
            i += 1
            return i
        self.assertEqual(test(1, 2, 3), 1)
        self.assertEqual(test(1, 2, 3), 1)
        self.assertEqual(test(1, 2, 3), 1)
        self.assertEqual(test.cache_info().hits, 2)
        self.assertEqual(test(1, 2, 4), 2)
        self.assertEqual(test.cache_info().misses, 2)


class TestHoneCache(unittest.TestCase):

    def test_default_finder(self):

        @hone_cache()
        def test(prefix):
            return set(x for x in {'foo', 'foobar', 'baz'}
                       if x.startswith(prefix))
        self.assertEqual(test('f'), {'foo', 'foobar'})
        self.assertEqual(test.cache_info().misses, 1)
        self.assertEqual(test('fo'), {'foo', 'foobar'})
        self.assertEqual(test.cache_info().misses, 1)
        self.assertEqual(test('foob'), {'foobar'})
        self.assertEqual(test.cache_info().misses, 1)
        self.assertEqual(test(''), {'foo', 'foobar', 'baz'})
        self.assertEqual(test.cache_info().misses, 2)
        self.assertEqual(test('ba'), {'baz'})
        self.assertEqual(test.cache_info().misses, 2)

    def test_tito(self):
        """ Type In Type Out """
        for xtype in list, set, tuple, frozenset:

            @hone_cache()
            def test(prefix):
                return xtype(x for x in ('foo', 'foobar', 'baz')
                             if x.startswith(prefix))
            self.assertEqual(test('f'), xtype(('foo', 'foobar')))
            self.assertEqual(test.cache_info().misses, 1)
            self.assertEqual(test('fo'), xtype(('foo', 'foobar')))
            self.assertEqual(test.cache_info().misses, 1)
            self.assertEqual(test('foob'), xtype(('foobar',)))
            self.assertEqual(test.cache_info().misses, 1)
            self.assertEqual(test(''), xtype(('foo', 'foobar', 'baz')))
            self.assertEqual(test.cache_info().misses, 2)
            self.assertEqual(test('ba'), xtype(('baz',)))
            self.assertEqual(test.cache_info().misses, 2)

    def test_path_finder_simple(self):
        tree = {
            "level1": {
                "level2": {
                    "leaf": 123
                }
            }
        }

        @hone_cache(refineby='container')
        def test(path):
            offt = tree
            for x in path:
                offt = offt[x]
            return offt
        self.assertEqual(test(('level1',)), tree['level1'])
        self.assertEqual(test.cache_info().misses, 1)
        self.assertEqual(test(('level1',)), tree['level1'])
        self.assertEqual(test.cache_info().hits, 1)
        self.assertEqual(test(('level1', 'level2')), tree['level1']['level2'])
        self.assertEqual(test.cache_info().partials, 1)

    def test_path_finder_mixed_type(self):
        tree = {
            "level1": [{
                "level2": {
                    "leaf": 123
                }
            }]
        }

        @hone_cache(refineby='container')
        def test(path):
            offt = tree
            for x in path:
                offt = offt[x]
            return offt
        self.assertEqual(test(('level1',)), tree['level1'])
        self.assertEqual(test.cache_info().misses, 1)
        self.assertEqual(test(('level1',)), tree['level1'])
        self.assertEqual(test.cache_info().hits, 1)
        self.assertEqual(test(('level1', 0, 'level2')),
                         tree['level1'][0]['level2'])
        self.assertEqual(test.cache_info().partials, 1)



"""
Table layout.
"""

import collections
import csv
import functools
import inspect
import itertools
import json
import math
import operator
import re
import shutil
import sys
import time
import warnings
from ..rendering import beststr, vtmlrender, VTMLBuffer


class RowsNotFound(ValueError):
    """ Similar to StopIteration but not specific to iterator protocol. """
    pass


class Table(object):
    """ A visual layout for row oriented data (like csv).  Most of the code
    here is dedicated to fitting the data as losslessly as possible onto a
    terminal screen.

    For detailed help see `__init__()` or use the helper function tabulate()
    for simple use cases.

    This class is also a context manager which is applicable for printing
    streams of data or when combined with output formats that require closing
    tags. """

    column_padding = 2
    column_align = 'left'  # {left,right,center}
    title_align = 'left'
    cliptext = beststr('…', '...')
    column_minwidth = len(cliptext)
    renderer_types = {}
    overflow_modes = 'clip', 'wrap', 'preformatted'

    def __init__(self, columns=None, headers=None, accessors=None, width=None,
                 clip=None, overflow=None, flex=True, file=None, cliptext=None,
                 column_minwidth=None, column_padding=None, column_align=None,
                 renderer=None, title=None, title_align=None, column_mask=None,
                 hide_header=False, hide_footer=False, align_rows=True,
                 justify=None):
        """ The .columns should be a list of width specs or a style dict.
        Width specs can be whole numbers representing fixed char widths,
        fractions representing percentages of available table width or None
        if the column is unspec'd.  Unspec'd columns are then calculated at
        render time according to the .flex setting.  If not provided then all
        columns will be unspec'd.  E.g.

            >>> t = Table(columns=[
                25, # fixed width, 25 chars
                50, # fixed width, 50 chars
                0.33, # 33% percent of total available space
                None, # if flex=True will size to row content
                None  # if flex=False will split remainder
            ])

        The columns style dict supports the following properties:

            "width":    Follows the aforementioned column width definition.
            "minwidth": Minimum width of column.
            "padding":  Property for custom padding of individual columns.
                        Whole number of white space characters to add.
            "align":    How to justify the contents of the column. Valid
                        choices are 'left', 'right', and 'center'.
            "overflow": How the column should be treated when it is too wide
                        to fit in the allotted space.  The default value comes
                        from the table's `overflow` global option.

            E.g.

            >>> t = Table(columns=[{
                "width": 1/4,
                "padding": 4,
                "align": "right"
            }, {
                "width": 1/4,
                "padding": 0,
                "align": "center"
            }, {
                "padding": 0,
                "overflow": "clip"
            }])

        When any values are omitted from the column def they will pickup
        default settings from the table wide .column_padding, .column_align
        and .column_minwidth attributes respectively.

        The .headers arg should contain a list of strings to display on the
        first line or be unset to skip the header line.

        The .accessors argument is to be used in conjunction with datasets
        are not 1:1 conforming lists.  Typically this means the data is some
        other type like a dictionary or an object.  The accessors list should
        be a list of strings (dictionary keys) or functions that can return
        the correct datum for each column respectively. E.g.

            >>> t = Table(headers=['First', 'Second', 'Last'],
                          accessors=['one', 'two', 'three'])
            >>> t.print_row({"three": "333", "one": "1", "two": "22"})

        If .flex is turned on and you have unspec columns (None) the render
        method will scan your initial data stream to determine the best widths
        to use.  It is permissible to submit your entire data set at this
        point if you have enough memory.

        To control overflow of wide columns set the .overflow argument to one
        of the .overflow_modes.  None will defer handling to the renderer class
        used for this table.  `clip` will shorten wide strings.  `wrap` will
        continue long lines on the next row affecting the entire table output.
        `preformatted` is used to ensure the columns width always fits the
        contents even if it requires the table to expand beyond the requested
        width.

        Note that .overflow can be set in the column specification as well,
        which will take precedence over this setting.

        Setting .justify to True will evenly distribute extra column space
        so the table fills the requested width.
        """
        if clip is not None:
            warnings.warn('clip is deprecated, use overflow=clip',
                          DeprecationWarning)
            if overflow is not None:
                raise TypeError('`clip` and `overflow` are mutually exclusive')
            elif clip:
                overflow = 'clip'
        if overflow is not None and overflow not in self.overflow_modes:
            raise TypeError("Invalid overflow mode: %s (choices: %s)" % (
                overflow, ', '.join(map(str, self.overflow_modes))))
        self.overflow = overflow
        self.title = title
        # Freeze the table definitions...
        try:
            self.columns_def = columns.copy() if columns is not None else None
        except AttributeError:
            self.columns_def = tuple(columns)
        self.accessors_def = tuple(accessors or ())
        self.headers = tuple(headers or ())
        self.width = width
        self.flex = flex
        self._file = file
        self.hide_header = hide_header
        self.hide_footer = hide_footer
        self.column_mask = column_mask
        self.align_rows = align_rows
        self.default_renderer = None
        self.renderer = renderer
        self.justify = justify
        if cliptext is not None:
            self.cliptext = cliptext
        if column_padding is not None:
            self.column_padding = column_padding
        if column_minwidth is not None:
            self.column_minwidth = column_minwidth
        if column_align is not None:
            self.column_align = column_align
        if title_align is not None:
            self.title_align = title_align

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return self.close(exception=exc)

    def close(self, exception=None):
        if self.default_renderer:
            self.default_renderer.close(exception=exception)

    @property
    def file(self):
        if self._file is not None:
            return self._file
        else:
            return sys.stdout

    def make_renderer(self):
        if not self.renderer:
            if not self.file.isatty():
                renderer = 'plain'
            else:
                renderer = 'terminal'
        else:
            renderer = self.renderer
        Renderer = self.renderer_types[renderer]
        return Renderer(self)

    @classmethod
    def register_renderer(cls, renderer):
        cls.renderer_types[renderer.name] = renderer

    @classmethod
    def unregister_renderer(cls, renderer):
        del cls.renderer_types[renderer.name]

    @classmethod
    def attach_arguments(cls, parser, prefix='--', skip_formats=False,
                         format_excludes=None, format_title=None,
                         format_desc=None, skip_render=False,
                         render_excludes=None, render_title=None,
                         render_desc=None, skip_filters=False,
                         filter_excludes=None, filter_title=None,
                         filter_desc=None):
        """ Attach argparse arguments to an argparse parser/group with table
        options.  These are renderer options and filtering options with the
        ability to turn off headers and footers.  The return value is function
        that parses an argparse.Namespace object into keyword arguments for a
        layout.Table constructor. """
        convs = []
        if not skip_formats:
            attach = cls.attach_format_arguments
            convs.append(attach(parser, prefix, format_excludes, format_title,
                                format_desc))
        if not skip_render:
            attach = cls.attach_render_arguments
            convs.append(attach(parser, prefix, render_excludes, render_title,
                                render_desc))
        if not skip_filters:
            attach = cls.attach_filter_arguments
            convs.append(attach(parser, prefix, filter_excludes, filter_title,
                                filter_desc))

        def argparse_ns_to_table_opts(ns):
            options = {}
            for conv in convs:
                options.update(conv(ns))
            return options
        return argparse_ns_to_table_opts

    @classmethod
    def attach_render_arguments(cls, parser, prefix='--', excludes=None,
                                title=None, desc=None):
        excludes = excludes or set()
        title = 'table render settings' if title is None else title
        desc = 'Overrides for table render settings.' if desc is None else desc
        group = parser.add_argument_group(title, description=desc)
        if 'overflow' not in excludes:
            group.add_argument('%soverflow' % prefix,
                               choices=cls.overflow_modes,
                               help='Override the default overflow behavior.')
        if 'table_width' not in excludes:
            group.add_argument('%stable-width' % prefix, type=int,
                               metavar='COLS',
                               help='Specify the table width in columns.')
        if 'column_padding' not in excludes:
            group.add_argument('%scolumn-padding' % prefix, type=int,
                               metavar='COLS',
                               help='Specify whitespace padding for each '
                               'table column in characters.')
        if 'column_align' not in excludes:
            group.add_argument('%scolumn-align' % prefix, metavar='JUSTIFY',
                               choices={'left', 'center', 'right'},
                               help='Table column justification.')

        def ns2table(ns):
            opts = {}
            if ns.overflow is not None:
                opts['overflow'] = ns.overflow
            if ns.table_width is not None:
                opts['width'] = ns.table_width
            if ns.column_padding is not None:
                opts['column_padding'] = ns.column_padding
            if ns.column_align:
                opts['column_align'] = ns.column_align
            return opts
        return ns2table

    @classmethod
    def attach_format_arguments(cls, parser, prefix='--', excludes=None,
                                title=None, desc=None):
        excludes = excludes or set()
        title = 'table output format' if title is None else title
        desc = 'Selection of output formats for table display.  The ' \
               'default behavior is to detect the output device\'s ' \
               'capabilities.' if desc is None else desc
        group = parser.add_argument_group(title, description=desc)
        ex_group = group.add_mutually_exclusive_group()
        for name, renderer in sorted(cls.renderer_types.items()):
            if name in excludes:
                continue
            ex_group.add_argument('%s%s' % (prefix, name), dest='table_format',
                                  action='store_const', const=name,
                                  help=inspect.getdoc(renderer))

        def ns2table(ns):
            return {
                "renderer": ns.table_format,
            }
        return ns2table

    @classmethod
    def attach_filter_arguments(cls, parser, prefix='--', excludes=None,
                                title=None, desc=None):
        title = 'table filters' if title is None else title
        desc = 'Options for filtering the table display.' if desc is None \
               else desc
        group = parser.add_argument_group(title, description=desc)
        excludes = excludes or set()
        if 'columns' not in excludes:
            group.add_argument('%scolumns' % prefix, dest='table_columns',
                               metavar="COL_INDEX", nargs='+', type=int,
                               help="Only show specific columns.")
        if 'no-header' not in excludes:
            group.add_argument('%sno-header' % prefix, dest='no_table_header',
                               action='store_true', help="Hide table header.")
        if 'no-footer' not in excludes:
            group.add_argument('%sno-footer' % prefix, dest='no_table_footer',
                               action='store_true', help="Hide table footer.")

        def ns2table(ns):
            return {
                "column_mask": ns.table_columns,
                "hide_header": ns.no_table_header,
                "hide_footer": ns.no_table_footer
            }
        return ns2table

    def make_accessors(self, columns):
        """ Accessors can be numeric keys for sequence row data, string keys
        for mapping row data, or a callable function.  For numeric and string
        accessors they can be inside a 2 element tuple where the 2nd value is
        the default value;  Similar to dict.get(lookup, default). """
        accessors = list(self.accessors_def or range(columns))
        for i, x in enumerate(accessors):
            if not callable(x):
                if isinstance(x, collections.abc.Sequence) and \
                   not isinstance(x, str):
                    key, default = x
                else:
                    key = x
                    default = ''

                def acc(row, key=key, default=default):
                    try:
                        return row[key]
                    except (KeyError, IndexError):
                        return default
                accessors[i] = acc
        return accessors

    def create_colspec(self, columns, overflow_default=None):
        """ Produce a full format columns spec dictionary.  The overrides spec
        can be a partial columns spec as described in the __init__ method's
        depiction of the .columns attribute. """
        spec = [{
            "width": None,
            "minwidth": self.column_minwidth,
            "padding": self.column_padding,
            "align": self.column_align,
            "overflow": self.overflow or overflow_default
        } for x in range(columns)]
        if self.columns_def:
            for dst, src in zip(spec, self.columns_def):
                if hasattr(src, 'items'):
                    dst.update(src)
                else:
                    dst['width'] = src
        return spec

    def column_mask_filter(self, items):
        if not self.column_mask:
            return items
        else:
            return [x for i, x in enumerate(items, 1) if i in self.column_mask]

    def print(self, rows):
        """ Write the data to our output stream (stdout).  If the table is not
        rendered yet, we will make a renderer instance which will freeze
        state. """
        if not self.default_renderer:
            self.default_renderer = self.make_renderer()
        self.default_renderer.print(rows)

    def print_row(self, row):
        return self.print([row])

    def print_footer(self, content):
        if self.hide_footer:
            return
        if not self.default_renderer:
            self.print([])
        self.default_renderer.print_footer_raw(content)


class TableRenderer(object):
    """ Interface for rendering table output to a screen, file or other data
    exchange. """

    name = None
    overflow_default = None

    def __init__(self, table):
        """ All calculated values required for rendering a table are kept
        here.  In theory a single Table instance can be used to render
        multiple and differing datasets in a concurrent system.  Admittedly
        this is over-engineered for a CLI suite and the result of a lazy
        Sunday. """
        self.colspec = None
        self.accessors = None
        self.headers = None
        self.headers_drawn = False
        self.footers_drawn = False
        self.filter_pipeline = None
        self.table = table

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close(exception=exc)

    def close(self, exception=None):
        pass

    def compute_style_filter(self, next_filter):
        t = self.table
        columns = (t.columns_def and len(t.columns_def)) or \
                  (t.headers and len(t.headers)) or \
                  (t.accessors_def and len(t.accessors_def))
        data = None
        if not columns:
            data = (yield)
            columns = len(data)
        self.accessors = t.column_mask_filter(t.make_accessors(columns))
        self.colspec = t.column_mask_filter(
            t.create_colspec(columns, overflow_default=self.overflow_default))
        self.headers = t.headers and t.column_mask_filter(t.headers[:])
        next(next_filter)
        if data is not None:
            next_filter.send(data)
        while True:
            next_filter.send((yield))

    def print_headers(self, headers):
        raise NotImplementedError()

    def print_title(self, title):
        raise NotImplementedError()

    def print_row(self, row):
        raise NotImplementedError()

    def print_footer(self, content):
        raise NotImplementedError()

    def print_footer_raw(self, raw_content):
        self.print_footer(self.cell_format(raw_content))

    def cell_format(self, value):
        """ Subclasses should put any visual formatting specific to their
        rendering type here. """
        return vtmlrender(value)

    def get_aligner(self, alignment, width):
        align_funcs = {
            "left": 'ljust',
            "right": 'rjust',
            "center": 'center'
        }
        return operator.methodcaller(align_funcs[alignment], width)

    def render_filter(self, next_filter):
        """ Produce formatted output from the raw data stream. """
        next(next_filter)
        while True:
            data = (yield)
            res = [self.cell_format(access(data)) for access in self.accessors]
            next_filter.send(res)

    def get_filters(self):
        """ Coroutine based filters for render pipeline. """
        return [
            self.compute_style_filter,
            self.render_filter
        ]

    def print(self, data):
        if self.filter_pipeline is None:
            pipeline = self.get_filters()
            head = self.printer()
            for fn in reversed(pipeline):
                head = fn(head)
            self.filter_pipeline = head
            next(head)
        for x in data:
            self.filter_pipeline.send(x)
        self.filter_pipeline.close()
        self.filter_pipeline = None

    def printer(self):
        if not self.table.hide_header and self.table.title:
            self.print_title(self.cell_format(self.table.title))
        while True:
            if not self.table.hide_header and \
               not self.headers_drawn and \
               any(self.headers):
                self.print_headers([self.cell_format(x or '')
                                    for x in self.headers])
                self.headers_drawn = True
            self.print_row((yield))


class VisualTableRenderer(TableRenderer):
    """ ABC used for renderers that draw visually. """

    overflow_default = 'preformatted'
    justify_default = False
    linebreak = beststr('—', '-')
    title_tpl = '\n<b>{:vtml}</b>\n'
    header_tpl = '<reverse>{:vtml}</reverse>'
    footer_tpl = '<dim>{:vtml}</dim>'
    default_width = 95

    # You probably shouldn't mess with these unless you really need custom
    # rendering performance.  Chances are you really don't and should
    # manage your data stream more carefully first.
    min_render_prefill = 5
    max_render_prefill = 1000
    max_render_delay = 2

    def __init__(self, table):
        super().__init__(table)
        self.width = None
        self.data_window = collections.deque(maxlen=self.max_render_prefill)

    @property
    def usable_width(self):
        """ The available combined character width when all padding is
        removed. """
        return self.width - sum(x['padding'] for x in self.colspec)

    @property
    def viewable_width(self):
        """ The available combined character width when all padding is
        removed. """
        return sum(self.widths) + sum(x['padding'] for x in self.colspec)

    @property
    def desired_width(self):
        if self.table.width is not None:
            return self.table.width
        elif self.table.file not in (sys.stdout, sys.stderr):
            return self.default_width
        else:
            return shutil.get_terminal_size()[0]

    def print_headers(self, headers):
        lines = [VTMLBuffer('').join(x) for x in self.format_row(headers)]
        for line in lines:
            print(self.cell_format(self.header_tpl.format(line)),
                  file=self.table.file)

    def print_title(self, title):
        title = self.title_tpl.format(self.format_fullwidth(title))
        print(self.cell_format(title), file=self.table.file)

    def print_row(self, row, rstrip=True):
        """ Format and print the pre-rendered data to the output device. """
        line = ''.join(map(str, row))
        print(line.rstrip() if rstrip else line, file=self.table.file)

    def print_footer(self, content):
        row = self.format_fullwidth(content)
        if not self.footers_drawn:
            self.footers_drawn = True
            self.print_linebreak()
        print(self.cell_format(self.footer_tpl.format(row)),
              file=self.table.file)

    def print_linebreak(self):
        print(self.linebreak * self.viewable_width, file=self.table.file)

    def format_row(self, row):
        """ Apply overflow, justification and padding to a row.  Returns lines
        (plural) of rendered text for the row. """
        assert all(isinstance(x, VTMLBuffer) for x in row)
        raw = (fn(x) for x, fn in zip(row, self.formatters))
        for line in itertools.zip_longest(*raw):
            line = list(line)
            for i, col in enumerate(line):
                if col is None:
                    line[i] = self._get_blank_cell(i)
            yield line

    def format_fullwidth(self, value):
        """ Return a full width column. Note that the padding is inherited
        from the first cell which inherits from column_padding. """
        assert isinstance(value, VTMLBuffer)
        pad = self.colspec[0]['padding']
        fmt = self.make_formatter(self.width - pad, pad,
                                  self.table.title_align)
        return VTMLBuffer('\n').join(fmt(value))

    def make_formatter(self, width, padding, alignment, overflow=None):
        """ Create formatter function that factors the width and alignment
        settings. """
        if overflow is None:
            overflow = self.overflow_default
        if overflow == 'clip':
            overflower = lambda x: [x.clip(width, self.table.cliptext)]
        elif overflow == 'wrap':
            overflower = lambda x: x.wrap(width)
        elif overflow == 'preformatted':
            overflower = lambda x: x.split('\n')
        else:
            raise RuntimeError("Unexpected overflow mode: %r" % overflow)
        align = self.get_aligner(alignment, width)
        pad = self.get_aligner('center', width + padding)
        return lambda value: [pad(align(x)) for x in overflower(value)]

    def make_formatters(self):
        """ Create a list formatter functions for each column.  They can then
        be stored in the render spec for faster justification processing. """
        return [self.make_formatter(inner_w, spec['padding'], spec['align'],
                                    spec['overflow'])
                for spec, inner_w in zip(self.colspec, self.widths)]

    def _uniform_dist(self, spread, total):
        """ Produce a uniform distribution of `total` across a list of
        `spread` size. The result is non-random and uniform. """
        fraction, fixed_increment = math.modf(total / spread)
        fixed_increment = int(fixed_increment)
        balance = 0
        dist = []
        for _ in range(spread):
            balance += fraction
            withdrawl = 1 if balance > 0.5 else 0
            if withdrawl:
                balance -= withdrawl
            dist.append(fixed_increment + withdrawl)
        return dist

    def get_filters(self):
        """ Coroutine based filters for render pipeline. """
        return [
            self.compute_style_filter,
            self.render_filter,
            self.calc_widths_filter,
            self.format_row_filter,
            self.align_rows_filter,
        ]

    def format_row_filter(self, next_filter):
        """ Apply overflow, justification, padding and expansion to a row. """
        next(next_filter)
        while True:
            items = (yield)
            assert all(isinstance(x, VTMLBuffer) for x in items)
            raw = (fn(x) for x, fn in zip(items, self.formatters))
            for x in itertools.zip_longest(*raw):
                next_filter.send(x)

    def align_rows_filter(self, next_filter):
        align_coro = self._column_pad_filter if self.table.align_rows else \
            self._column_pack_filter
        aligner = align_coro(next_filter)
        next(aligner)
        while True:
            aligner.send((yield))

    def _column_pack_filter(self, next_filter):
        """ Top-align column data irrespective of original row alignment.  E.g.
            INPUT: [
                ["1a", "2a"],
                [None, "2b"],
                ["1b", "2c"],
                [None, "2d"]
            ]
            OUTPUT: [
                ["1a", "2a"],
                ["1b", "2b"],
                [<blank>, "2c"],
                [<blank>, "2d"]
            ]
        """
        next(next_filter)
        col_count = len(self.widths)
        queues = [collections.deque() for _ in range(col_count)]
        while True:
            try:
                row = (yield)
            except GeneratorExit:
                break
            for col, queue in zip(row, queues):
                if col is not None:
                    queue.append(col)
            if all(queues):
                next_filter.send([x.popleft() for x in queues])
        blanks = list(map(self._get_blank_cell, range(col_count)))
        while any(queues):
            next_filter.send([q.popleft() if q else blank
                              for q, blank in zip(queues, blanks)])

    def _column_pad_filter(self, next_filter):
        """ Expand blank lines caused from overflow of other columns to blank
        whitespace.  E.g.
            INPUT: [
                ["1a", "2a"],
                [None, "2b"],
                ["1b", "2c"],
                [None, "2d"]
            ]
            OUTPUT: [
                ["1a", "2a"],
                [<blank>, "2b"],
                ["1b", "2c"],
                [<blank>, "2d"]
            ]
        """
        next(next_filter)
        while True:
            line = list((yield))
            for i, col in enumerate(line):
                if col is None:
                    line[i] = self._get_blank_cell(i)
            next_filter.send(line)

    @functools.lru_cache()
    def _get_blank_cell(self, index):
        """ Return a formatted blank cell for a specific column index. """
        return self.formatters[index](VTMLBuffer())[0]

    def width_normalize(self, width):
        """ Handle a width style, which can be a fractional number
        representing a percentage of available width or positive integers
        which indicate a fixed width. """
        if width is not None:
            if width > 0 and width < 1:
                return int(width * self.usable_width)
            else:
                return int(width)

    def calc_widths_filter(self, next_filter):
        """ Coroutine to analyze the incoming data stream for creating optimal
        column width choices.  This may buffer some of the incoming stream if
        there isn't enough information to make good choices about column
        widths.  Also it may resize widths if certain conditions are met such
        as the terminal width resize event being detected. """
        window_sent = not not self.data_window
        next_primed = False
        genexit = None
        if not self.data_window:
            start = time.monotonic()
            while len(self.data_window) < self.min_render_prefill or \
                (len(self.data_window) < self.max_render_prefill and
                 (time.monotonic() - start) < self.max_render_delay):
                try:
                    self.data_window.append((yield))
                except GeneratorExit as e:
                    genexit = e
                    break
        while True:
            if self.width != self.desired_width:
                self.headers_drawn = False  # TODO: make optional
                self.width = self.desired_width
                remaining = self.usable_width
                widths = [x['width'] for x in self.colspec]
                preformatted = [i for i, x in enumerate(self.colspec)
                                if x['overflow'] == 'preformatted']
                unspec = []
                for i, width in enumerate(widths):
                    fixed_width = self.width_normalize(width)
                    if fixed_width is None:
                        unspec.append(i)
                    else:
                        widths[i] = fixed_width
                        remaining -= fixed_width
                if unspec:
                    if self.table.flex and self.data_window:
                        for i, w in self.calc_flex(self.data_window, remaining,
                                                   unspec, preformatted):
                            widths[i] = w
                    else:
                        dist = self._uniform_dist(len(unspec), remaining)
                        for i, width in zip(unspec, dist):
                            widths[i] = width
                self.widths = widths
                self.formatters = self.make_formatters()
            if not next_primed:
                next(next_filter)
                next_primed = True
            if not window_sent:
                for x in self.data_window:
                    next_filter.send(x)
                window_sent = True
            if genexit:
                raise genexit
            data = (yield)
            self.data_window.append(data)
            next_filter.send(data)

    def calc_flex(self, data, max_width, cols, preformatted=None):
        """ Scan data returning the best width for each column given the
        max_width constraint.  If some columns will overflow we calculate the
        best concession widths. """
        if preformatted is None:
            preformatted = []
        colstats = []
        for i in cols:
            lengths = [len(xx) for x in data
                       for xx in x[i].text().splitlines()]
            if self.headers:
                lengths.append(len(self.headers[i]))
            lengths.append(self.width_normalize(self.colspec[i]['minwidth']))
            counts = collections.Counter(lengths)
            colstats.append({
                "column": i,
                "preformatted": i in preformatted,
                "counts": counts,
                "offt": max(lengths),
                "chop_mass": 0,
                "chop_count": 0,
                "total_mass": sum(a * b for a, b in counts.items())
            })
        self.adjust_widths(max_width, colstats)
        required = sum(x['offt'] for x in colstats)
        justify = self.table.justify if self.table.justify is not None else \
            self.justify_default
        if required < max_width and justify:
            # Fill remaining space proportionately.
            remaining = max_width
            for x in colstats:
                x['offt'] = int((x['offt'] / required) * max_width)
                remaining -= x['offt']
            if remaining:
                dist = self._uniform_dist(len(cols), remaining)
                for adj, col in zip(dist, colstats):
                    col['offt'] += adj
        return [(x['column'], x['offt']) for x in colstats]

    def adjust_widths(self, max_width, colstats):
        """ Adjust column widths based on the least negative affect it will
        have on the viewing experience.  We take note of the total character
        mass that will be clipped when each column should be narrowed.  The
        actual score for clipping is based on percentage of total character
        mass, which is the total number of characters in the column. """
        adj_colstats = []
        for x in colstats:
            if not x['preformatted']:
                adj_colstats.append(x)
            else:
                max_width -= x['offt']
        next_score = lambda x: (x['counts'][x['offt']] + x['chop_mass'] +
                                x['chop_count']) / x['total_mass']
        cur_width = lambda: sum(x['offt'] for x in adj_colstats)
        min_width = lambda x: self.width_normalize(
            self.colspec[x['column']]['minwidth'])
        while cur_width() > max_width:
            nextaffects = [(next_score(x), i)
                           for i, x in enumerate(adj_colstats)
                           if x['offt'] > min_width(x)]
            if not nextaffects:
                break  # All columns are as small as they can get.
            nextaffects.sort()
            chop = adj_colstats[nextaffects[0][1]]
            chop['chop_count'] += chop['counts'][chop['offt']]
            chop['chop_mass'] += chop['chop_count']
            chop['offt'] -= 1


class PlainTableRenderer(VisualTableRenderer):
    """ Render output without any vt100 codes. """

    name = 'plain'

    def cell_format(self, value):
        return vtmlrender(value, plain=True)

    def print_headers(self, headers):
        super().print_headers(headers)
        self.print_linebreak()

Table.register_renderer(PlainTableRenderer)


class TerminalTableRenderer(VisualTableRenderer):
    """ Render a table designed to fit/fill a terminal.  This renderer produces
    the most human friendly output when on a terminal device. """

    name = 'terminal'
    overflow_default = 'clip'
    justify_default = True

Table.register_renderer(TerminalTableRenderer)


class DataTableRenderer(TableRenderer):
    """ ABC intended for data renderers. """

    def cell_format(self, value):
        return vtmlrender(value, plain=True)


class JSONTableRenderer(DataTableRenderer):
    """ Generate JSON output of the table. """

    name = 'json'
    key_split = re.compile('[\s\-_\.\/]')
    key_filter = re.compile('[^a-zA-Z0-9]')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.seen_keys = set()
        self.keys = []
        self.buf = {
            'title': None,
            'rows': [],
            'footers': []
        }

    def make_key(self, value):
        """ Make camelCase variant of value. """
        if value:
            parts = [self.key_filter.sub('', x)
                     for x in self.key_split.split(value.lower())]
            key = parts[0] + ''.join(map(str.capitalize, parts[1:]))
        else:
            key = ''
        if key in self.seen_keys:
            i = 1
            while '%s%d' % (key, i) in self.seen_keys:
                i += 1
            key = '%s%d' % (key, i)
        self.seen_keys.add(key)
        return key

    def print_title(self, title):
        self.buf['title'] = title.text()

    def print_headers(self, headers):
        self.keys[:] = map(self.make_key, [x.text() for x in headers])

    def print_footer(self, content):
        self.buf['footers'].append(self.cell_format(content).text())

    def print_row(self, row):
        row = [x.text() for x in row]
        self.buf['rows'].append(dict(zip(self.keys, row)))

    def close(self, exception=None):
        if exception and any(exception):
            return
        print(json.dumps(self.buf, indent=4, sort_keys=True),
              file=self.table.file)

Table.register_renderer(JSONTableRenderer)


class CSVTableRenderer(DataTableRenderer):
    """ Generate CSV (comma delimited) output of the table. """

    name = 'csv'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.writer = csv.writer(self.table.file)

    def print_title(self, title):
        """ CSV does not support a title. """
        pass

    def print_headers(self, headers):
        self.writer.writerow(headers)

    def print_row(self, row):
        self.writer.writerow(row)

    def print_footer(self, content):
        """ CSV does not support footers. """
        pass

Table.register_renderer(CSVTableRenderer)


class MarkdownTableRenderer(PlainTableRenderer):
    """ Render Markdown text format output of the table. """

    name = 'md'

    @property
    def usable_width(self):
        """ Allocate space for markdown borders. """
        border_width = len(self.colspec) + 1
        return super().usable_width - border_width

    def mdprint(self, *columns):
        print('|%s|' % '|'.join(map(str, columns)), file=self.table.file)

    def print_title(self, title):
        print("\n**%s**\n" % self.format_fullwidth(title).text(),
              file=self.table.file)

    def print_headers(self, headers):
        for line in self.format_row(headers):
            self.mdprint(*map(str, line))
        self.mdprint(*("-" * (width + colspec['padding'])
                     for width, colspec in zip(self.widths, self.colspec)))

    def print_row(self, row):
        self.mdprint(*row)

    def print_footer(self, content):
        print("\n_%s_" % content, file=self.table.file)

Table.register_renderer(MarkdownTableRenderer)


def tabulate(data, header=True, headers=None, accessors=None,
             **table_options):
    """ Shortcut function to produce tabular output of data without the
    need to create and configure a Table instance directly. The function
    does however return a table instance when it's done for any further use
    by the user. """
    if header and not headers:
        data = iter(data)
        try:
            headers = next(data)
        except StopIteration:
            pass
    if headers and hasattr(headers, 'items') and accessors is None:
        # Dict mode; Build accessors and headers from keys of data.
        data = itertools.chain([headers], data)
        accessors = list(headers)
        headers = [' '.join(map(str.capitalize, x.replace('_', ' ').split()))
                   for x in accessors]
    t = Table(headers=headers, accessors=accessors, **table_options)
    try:
        t.print(data)
    except RowsNotFound:
        pass
    return t



"""
Tree layout.
"""

import collections
import collections.abc
import sys
from .. import rendering


class TreeNode(object):

    def __init__(self, value, children=None, label=None):
        self.value = value
        self.label = label
        self.children = children if children is not None else []

    def __lt__(self, item):
        return self.value < item.value


class Tree(object):
    """ Construct a visual tree from a data source. """

    tree_L = rendering.beststr('└── ', '\-- ')
    tree_T = rendering.beststr('├── ', '+-- ')
    tree_vertspace = rendering.beststr('│   ', '|   ')

    def __init__(self, formatter=None, sort_key=None):
        self.formatter = formatter or self.default_formatter
        self.sort_key = sort_key

    def default_formatter(self, node):
        if node.label is not None:
            return '%s: <b>%s</b>' % (node.value, node.label)
        else:
            return str(node.value)

    def render(self, nodes, prefix=None):
        node_list = list(nodes)
        end = len(node_list) - 1
        if self.sort_key is not False:
            node_list.sort(key=self.sort_key)
        for i, x in enumerate(node_list):
            if prefix is not None:
                line = [prefix]
                if end == i:
                    line.append(self.tree_L)
                else:
                    line.append(self.tree_T)
            else:
                line = ['']
            yield rendering.vtmlrender(''.join(line + [self.formatter(x)]))
            if x.children:
                if prefix is not None:
                    line[-1] = '    ' if end == i else self.tree_vertspace
                yield from self.render(x.children, prefix=''.join(line))


def treeprint(data, render_only=False, file=None, **options):
    """ Render a tree structure based on generic python containers. The keys
    should be titles and the values are children of the node or None if it's
    an empty leaf node;  Primitives are valid leaf node labels too.  E.g.

        sample = {
            "Leaf 1": None,
            "Leaf 2": "I have a label on me",
            "Branch A": {
                "Sub Leaf 1 with float label": 3.14,
                "Sub Branch": {
                    "Deep Leaf": None
                }
            },
            "Branch B": {
                "Sub Leaf 2": None
            }
        }
    """

    def getiter(obj):
        if isinstance(obj, collections.abc.Mapping):
            return obj.items()
        elif (isinstance(obj, collections.abc.Iterable) and
              not isinstance(obj, str)):
            return enumerate(obj)

    def cycle_check(item, seen=set()):
        item_id = id(item)
        if item_id in seen:
            raise ValueError('Cycle detected for: %s' % repr(item))
        else:
            seen.add(item_id)

    def crawl(obj, cc=cycle_check):
        cc(obj)
        objiter = getiter(obj)
        if objiter is None:
            yield TreeNode(obj)
        else:
            for key, item in objiter:
                if isinstance(item, collections.abc.Iterable) and \
                   not isinstance(item, str):
                    yield TreeNode(key, children=crawl(item))
                elif item is None:
                    yield TreeNode(key)
                else:
                    yield TreeNode(key, label=item)
    t = Tree(**options)
    render_gen = t.render(crawl(data))
    if render_only:
        return render_gen
    else:
        file = sys.stdout if file is None else file
        conv = (lambda x: x.plain()) if not file.isatty() else (lambda x: x)
        for x in render_gen:
            print(conv(x), file=file)



import argparse
import csv
import shellish
import sys

csvfile = argparse.FileType('r', encoding='utf-8-sig')


@shellish.autocommand
def csvpretty(csvfile: csvfile=sys.stdin):
    """ Pretty print a CSV file. """
    shellish.tabulate(csv.reader(csvfile))



"""
VTML traceback/exception formatting and printing

Many of the methods take from the naming convention of the standard library's
traceback module.
"""

import traceback
from . import vtml


def format_exception(exc, indent=0, pad='  '):
    """ Take an exception object and return a generator with vtml formatted
    exception traceback lines. """
    from_msg = None
    if exc.__cause__ is not None:
        indent += yield from format_exception(exc.__cause__, indent)
        from_msg = traceback._cause_message.strip()
    elif exc.__context__ is not None and not exc.__suppress_context__:
        indent += yield from format_exception(exc.__context__, indent)
        from_msg = traceback._context_message.strip()
    padding = pad * indent
    if from_msg:
        yield '\n%s%s\n' % (padding, from_msg)
    yield '%s<b><u>Traceback (most recent call last)</u></b>' % padding
    tblist = traceback.extract_tb(exc.__traceback__)
    tbdepth = len(tblist)
    for x in tblist:
        depth = '%d.' % tbdepth
        yield '%s<dim>%-3s</dim> <cyan>File</cyan> "<blue>%s</blue>", ' \
            'line <u>%d</u>, in <b>%s</b>' % (padding, depth, x.filename,
                                              x.lineno, x.name)
        yield '%s      %s' % (padding, x.line)
        tbdepth -= 1
    yield '%s<b><red>%s</red>: %s</b>' % (padding, type(exc).__name__, exc)
    return indent + 1


def print_exception(*args, file=None, **kwargs):
    """ Print the formatted output of an exception object. """
    for line in format_exception(*args, **kwargs):
        vtml.vtmlprint(line, file=file)



"""
Display a list in columns.
"""

import math
import shutil
import sys
from .. import rendering


def columnize(items, width=None, file=sys.stdout):
    """ Smart display width handling when showing a list of stuff. """
    if not items:
        return
    if width is None:
        width = shutil.get_terminal_size()[0] if file is sys.stdout else 80
    items = [rendering.vtmlrender(x) for x in items]
    maxcol = max(items, key=len)
    colsize = len(maxcol) + 2
    cols = width // colsize
    if cols < 2:
        for x in items:
            print(x, file=file)
        return
    lines = math.ceil(len(items) / cols)
    for i in range(lines):
        row = items[i:None:lines]
        print(*[x.ljust(colsize) for x in row], sep='', file=file)



"""
Demo of different ways to skin a cat.
"""

import shellish


##############
# Decorators #
##############
@shellish.autocommand
def cat1():
    """ Auto command cannot interact. """
    pass


@shellish.autocommand
def sub1(optional: int=1):
    print("ran subcommand1", optional)

cat1.add_subcommand(sub1)


###############
# Composition #
###############
cat2 = shellish.Command(name='cat2', title='composition cat')

sub2 = shellish.Command(name='sub2', title='composition cat sub')
sub2.add_argument('--optional', type=int, default=2)
sub2.run = lambda args: print("ran subcommand2", args.optional)
cat2.add_subcommand(sub2)


###############
# Inheritance #
###############
class Cat3(shellish.Command):
    """ Inheritance cat. """

    name = 'cat3'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_subcommand(Sub3)


class Sub3(shellish.Command):
    """ Inheritance cat sub. """

    name = 'sub3'

    def setup_args(self, parser):
        self.add_argument('--optional', type=int, default=3)

    def run(self, args):
        print("ran subcommand3", args.optional)


# Putting it together for a demo..
class Main(shellish.Command):

    def run(self, args):
        self.session.run_loop()


main = Main(name='main', title='harness')
main.add_subcommand(cat1)
main.add_subcommand(cat2)
main.add_subcommand(Cat3)
main()



"""
Command for interactive sessions only.
"""

from .. import command


class Exit(command.Command):
    """ Exit an interactive session. """

    name = 'exit'

    def run(self, args):
        self.session.exit()


class Reset(command.Command):
    """ Reset the TTY. """

    name = 'reset'

    def run(self, args):
        print('\033c')


class Pager(command.Command):
    """ Enable or disable the pager for commands. """

    name = 'pager'

    def enable(self, args):
        if self.session.allow_pager:
            raise SystemExit("Already enabled")
        self.session.allow_pager = True

    def disable(self, args):
        if not self.session.allow_pager:
            raise SystemExit("Already disabled")
        self.session.allow_pager = False

    def setup_args(self, parser):
        self.add_subcommand(command.Command(name='enable', run=self.enable))
        self.add_subcommand(command.Command(name='disable', run=self.disable))

    def run(self, args):
        if self.session.allow_pager:
            self.session.allow_pager = False
            print("Disabled pager")
        else:
            self.session.allow_pager = True
            print("Enabled pager: %s" % self.get_config('core').get('pager'))




import atexit
import sys


def linebuffered_stdout():
    """ Always line buffer stdout so pipes and redirects are CLI friendly. """
    if sys.stdout.line_buffering:
        return sys.stdout
    orig = sys.stdout
    new = type(orig)(orig.buffer, encoding=orig.encoding, errors=orig.errors,
                     line_buffering=True)
    new.mode = orig.mode
    return new

sys.stdout = linebuffered_stdout()


def ignore_broken_pipe():
    """ If a shellish program has redirected stdio it is subject to erroneous
    "ignored" exceptions during the interpretor shutdown. This essentially
    beats the interpretor to the punch by closing them early and ignoring any
    broken pipe exceptions. """
    for f in sys.stdin, sys.stdout, sys.stderr:
        try:
            f.close()
        except BrokenPipeError:
            pass

atexit.register(ignore_broken_pipe)


from .command import *  # noqa
from .autocommand import *  # noqa



import shellish
import unittest
import os


class EnvTests(unittest.TestCase):

    def setUp(self):
        self.env_save = env = os.environ
        os.environ = env.copy()

    def tearDown(self):
        os.environ = self.env_save

    def setenv(self, key, value):
        os.environ[key] = value

    def test_simple_env_noset(self):
        c = shellish.Command(name='c')
        c.add_argument('--foo', env='FOO')
        c.run = lambda args: args
        self.assertEqual(c(argv='--foo bar').foo, 'bar')
        self.assertEqual(c(argv='').foo, None)

    def test_dup_env(self):
        c = shellish.Command(name='c')
        c.add_argument('--foo', env='FOO')
        self.assertRaises(ValueError, c.add_argument, '--bar', env='FOO')

    def test_simple_env_set(self):
        c = shellish.Command(name='c')
        c.add_argument('--foo', env='SHELLISH_TEST_FOO')
        c.run = lambda args: args
        self.setenv('SHELLISH_TEST_FOO', 'from_env')
        self.assertEqual(c(argv='--foo from_args').foo, 'from_args')
        self.assertEqual(c(argv='').foo, 'from_env')

    def test_override_default(self):
        c = shellish.Command(name='c')
        c.add_argument('--foo', default='foodef', env='SHELLISH_TEST_FOO')
        c.run = lambda args: args
        self.assertEqual(c(argv='').foo, 'foodef')
        self.setenv('SHELLISH_TEST_FOO', 'from_env')
        self.assertEqual(c(argv='').foo, 'from_env')
        self.assertEqual(c(argv='--foo from_arg').foo, 'from_arg')

    def test_autoenv(self):
        c = shellish.Command(name="test_autoenv")
        c.add_argument('--foo', autoenv=True)
        c.run = lambda args: args  # just return args ns
        self.assertEqual(c(argv='').foo, None)
        self.assertEqual(c(argv='--foo bar').foo, 'bar')
        self.setenv('TEST_AUTOENV_FOO', 'from_env')
        self.assertEqual(c(argv='').foo, 'from_env')
        self.assertEqual(c(argv='--foo from_arg').foo, 'from_arg')

    def test_autoenv_nested_commands(self):
        head = shellish.Command(name="head")
        tail = shellish.Command(name="tail")
        tail.add_argument('--foo', autoenv=True)
        head.add_subcommand(tail)
        tail.run = lambda args: args  # just return args ns
        self.assertEqual(head(argv='tail').foo, None)
        self.assertEqual(head(argv='tail --foo bar').foo, 'bar')
        self.setenv('HEAD_TAIL_FOO', 'from_env')
        self.assertEqual(head(argv='tail').foo, 'from_env')
        self.assertEqual(head(argv='tail --foo from_arg').foo, 'from_arg')

    def test_autoenv_double_nested_commands(self):
        head = shellish.Command(name="head")
        mid = shellish.Command(name="mid")
        tail = shellish.Command(name="tail")
        tail.add_argument('--foo', autoenv=True)
        mid.add_subcommand(tail)
        head.add_subcommand(mid)
        tail.run = lambda args: args  # just return args ns
        self.assertEqual(head(argv='mid tail').foo, None)
        self.assertEqual(head(argv='mid tail --foo bar').foo, 'bar')
        self.setenv('HEAD_MID_TAIL_FOO', 'from_env')
        self.assertEqual(head(argv='mid tail').foo, 'from_env')
        self.assertEqual(head(argv='mid tail --foo from_arg').foo, 'from_arg')

    def test_autoenv_double_nested_commands_alt_order(self):
        head = shellish.Command(name="head")
        mid = shellish.Command(name="mid")
        head.add_subcommand(mid)
        tail = shellish.Command(name="tail")
        mid.add_subcommand(tail)
        tail.add_argument('--foo', autoenv=True)
        tail.run = lambda args: args  # just return args ns
        self.assertEqual(head(argv='mid tail').foo, None)
        self.assertEqual(head(argv='mid tail --foo bar').foo, 'bar')
        self.setenv('HEAD_MID_TAIL_FOO', 'from_env')
        self.assertEqual(head(argv='mid tail').foo, 'from_env')
        self.assertEqual(head(argv='mid tail --foo from_arg').foo, 'from_arg')

    def test_autoenv_odd_names(self):
        names = {
            "test-autoenv": "TEST_AUTOENV",
            "test ": "TEST",
            "test  ": "TEST",
            " test  ": "TEST",
            " test": "TEST",
            "  test": "TEST",
            "  test-test": "TEST_TEST",
            "  test test": "TEST_TEST",
            "  t!@#$est-test": "TEST_TEST",
            "1test": "_1TEST",
            " 1test": "_1TEST",
            " 1test ": "_1TEST",
        }
        for i, (name, env_prefix) in enumerate(names.items()):
            with self.subTest(name):
                c = shellish.Command(name=name)
                action = c.add_argument('--foo', autoenv=True)
                self.assertEqual(action.env, '%s_FOO' % env_prefix)

    def test_nonstd_parser(self):
        c = shellish.Command(name="test_autoenv")
        g = c.argparser.add_mutually_exclusive_group()
        c.add_argument('--foo', parser=g, env='SHELLISH_TEST_FOO')
        c.run = lambda args: args
        self.assertEqual(c(argv='').foo, None)
        self.assertEqual(c(argv='--foo bar').foo, 'bar')
        self.setenv('SHELLISH_TEST_FOO', 'from_env')
        self.assertEqual(c(argv='').foo, 'from_env')
        self.assertEqual(c(argv='--foo from_arg').foo, 'from_arg')



"""
A very simple event framework that can be used to more dynamically mixin
behavior for shellish activity.
"""


class Eventer(object):
    """ Very simple event mix-in. """

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._events = {}
        return instance

    def add_events(self, adding):
        """ Setup events for this instance.  Just a list of strings. """
        for x in adding:
            self._events.setdefault(x, [])

    def add_listener(self, event, callback, single=False, priority=1):
        """ Add a callback to an event list so it will be run at this event's
        firing. If single is True, the event is auto-removed after the first
        invocation. Priority can be used to jump ahead or behind other
        callback invocations."""
        event_stack = self._events[event]
        event_stack.append({
            "callback": callback,
            "single": single,
            "priority": priority
        })
        event_stack.sort(key=lambda x: x['priority'])

    def remove_listener(self, event, callback, single=None, priority=None):
        """ Remove the event listener matching the same signature used for
        adding it.  This will remove AT MOST one entry meeting the signature
        requirements. """
        event_stack = self._events[event]
        for x in event_stack:
            if x['callback'] == callback and \
               (single is None or x['single'] == single) and \
               (priority is None or x['priority'] == priority):
                event_stack.remove(x)
                break
        else:
            raise KeyError('Listener not found for "%s": %s' % (event,
                           callback))

    def fire_event(self, event, *args, **kwargs):
        """ Execute the listeners for this event passing any arguments
        along. """
        remove = []
        event_stack = self._events[event]
        for x in event_stack:
            x['callback'](*args, **kwargs)
            if x['single']:
                remove.append(x)
        for x in remove:
            event_stack.remove(x)



"""
A logging handler that's tty aware.
"""

import logging
from . import rendering


class VTMLHandler(logging.StreamHandler):
    """ Parse VTML messages to colorize and embolden logs. """

    def __init__(self, *args, fmt=None, level_prefmt=None, field_prefmt=None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.setFormatter(VTMLFormatter(fmt=fmt, level_prefmt=level_prefmt,
                                        field_prefmt=field_prefmt))

    def format(self, record):
        return str(rendering.vtmlrender(super().format(record)))


class VTMLFormatter(logging.Formatter):

    default_fmt = ' '.join((
        '[%(asctime)s]',
        '[%(name)s]',
        '[%(levelname)s]',
        '%(message)s'
    ))
    default_level_prefmt = {
        logging.DEBUG: '<dim>%s</dim>',
        logging.INFO: '%s',
        logging.WARNING: '<b>%s</b>',
        logging.ERROR: '<red>%s</red>',
        logging.CRITICAL: '<red><b>%s</b></red>',
    }
    default_field_prefmt = {
        "asctime": '<blue>%s</blue>',
        "created": '<blue>%s</blue>',
        "filename": '<magenta>%s</magenta>',
        "funcName": '<yellow>%s</yellow>',
        "lineno": '<cyan>%s</cyan>',
        "name": '<green>%s</green>',
        "process": '<b>%s</b>',
    }

    def __init__(self, fmt=None, field_prefmt=None, level_prefmt=None,
                 **kwargs):
        fmt = fmt or self.default_fmt
        field_prefmt = field_prefmt or self.default_field_prefmt
        self.field_prefmt = dict((k, v) for k, v in field_prefmt.items()
                                 if '%%(%s)' % k in fmt)
        self.level_prefmt = level_prefmt or self.default_level_prefmt
        self.asctime_prefmt = self.field_prefmt.pop('asctime', None)
        super().__init__(fmt=fmt, **kwargs)

    def formatException(self, ei):
        return '\n'.join(rendering.format_exception(ei[1]))

    def formatTime(self, *args, **kwargs):
        s = super().formatTime(*args, **kwargs)
        return self.asctime_prefmt % s

    def usesTime(self):
        return self.asctime_prefmt

    def format(self, record):
        record.levelname = self.level_prefmt[record.levelno] % record.levelname
        for key, fmt in self.field_prefmt.items():
            setattr(record, key, fmt % getattr(record, key))
        return super().format(record)



"""
Translate HTML (simple) to VTML so it can be printed to a terminal or saved
to a text file in human readable format.
"""

import collections
import re
import warnings
from . import vtml
from html import parser


class HTMLConv(parser.HTMLParser):
    """ Convert HTML to VTML. """

    strip = {
        "script",
        "style"
    }

    raw = {
        "code"
    }

    noop = {
        "head",
        "html",
        "div",
        "span",
        "pre"
    }

    whitespace = re.compile(r'\s+')

    def reset(self):
        self.buf = []
        self.tag_stack = []
        self.tag_attrs = collections.defaultdict(list)
        super().reset()

    def stripping(self):
        for x in self.strip:
            if x in self.tag_stack:
                return True
        return False

    def preserve_whitespace(self):
        for x in self.raw:
            if x in self.tag_stack:
                return True
        return False

    def handle_starttag(self, tag, attrs):
        if tag in self.noop:
            return
        self.tag_stack.append(tag)
        if tag not in self.strip:
            handler = getattr(self, 'handle_start_%s' % tag, None)
            if handler:
                attrs = collections.OrderedDict(attrs)
                self.tag_attrs[tag].append(attrs)
                handler(tag, attrs)
            else:
                warnings.warn('unhandled tag: %s' % tag)

    def handle_endtag(self, tag):
        if tag in self.noop:
            return
        if self.tag_stack[-1] != tag:
            warnings.warn("Bad close tag: %s; Expected: %s" % (tag,
                          self.tag_stack[-1]))
            return
        del self.tag_stack[-1]
        if tag not in self.strip:
            handler = getattr(self, 'handle_end_%s' % tag, None)
            if handler:
                attrs = self.tag_attrs[tag].pop()
                handler(tag, attrs)

    def handle_data(self, data):
        if not self.stripping():
            if not self.preserve_whitespace():
                data = self.whitespace.sub(' ', data)
            self.buf.append(data)

    def handle_start_b(self, tag, attrs):
        self.buf.append('<b>')
    handle_start_strong = handle_start_b

    def handle_end_b(self, tag, attrs):
        self.buf.append('</b>')
    handle_end_strong = handle_end_b

    def handle_start_i(self, tag, attrs):
        self.buf.append('<i>')

    def handle_end_i(self, tag, attrs):
        self.buf.append('</i>')

    def handle_start_u(self, tag, attrs):
        self.buf.append('<u>')
    handle_start_em = handle_start_u

    def handle_end_u(self, tag, attrs):
        self.buf.append('</u>')
    handle_end_em = handle_end_u

    def handle_start_h1(self, tag, attrs):
        self.buf.append('\n<b>')
    handle_start_h2 = handle_start_h1
    handle_start_h3 = handle_start_h1
    handle_start_h4 = handle_start_h1

    def handle_end_h1(self, tag, attrs):
        self.buf.append('</b>\n\n')
    handle_end_h2 = handle_end_h1
    handle_end_h3 = handle_end_h1
    handle_end_h4 = handle_end_h1

    def handle_start_small(self, tag, attrs):
        self.buf.append('<dim>')

    def handle_end_small(self, tag, attrs):
        self.buf.append('<dim>')

    def handle_start_a(self, tag, attrs):
        self.buf.append('<blue><u>')

    def handle_end_a(self, tag, attrs):
        href = attrs.get('href')
        if href:
            self.buf.append(' (%s)' % href)
        self.buf.append('</u></blue>')

    def handle_start_p(self, tag, attrs):
        self.buf.append('\n')

    def handle_end_p(self, tag, attrs):
        self.buf.append('\n')

    def handle_start_br(self, tag, attrs):
        self.buf.append('\n')

    def handle_end_br(self, tag, attrs):
        pass

    def handle_start_ol(self, tag, attrs):
        self.buf.append('\n')

    def handle_end_ol(self, tag, attrs):
        self.buf.append('\n')

    def handle_start_ul(self, tag, attrs):
        self.buf.append('\n')

    def handle_end_ul(self, tag, attrs):
        self.buf.append('\n')

    def handle_start_li(self, tag, attrs):
        for x in reversed(self.tag_stack):
            if x in ('ul', 'ol'):
                list_attrs = self.tag_attrs[x][-1]
                tag = x
                break
        else:
            warnings.warn("Bad LI tag is outside UL or OL")
        if tag == 'ol':
            bullet = '%d.' % list_attrs.setdefault('_index', 1)
            list_attrs['_index'] += 1
        else:
            bullet = vtml.beststr('●', '*')
        self.buf.append('<b> %s  </b>' % bullet)

    def handle_end_li(self, tag, attrs):
        self.buf.append('\n')

    def handle_start_code(self, tag, attrs):
        pass

    def handle_end_code(self, tag, attrs):
        lines = self.buf.pop().splitlines()
        width = len(max(lines, key=len))
        fmt = '  <bgblack><green>  %%-%ds  </green></bgblack>\n' % width
        self.buf.append('\n')
        self.buf.append(fmt % '')
        for line in lines:
            self.buf.append(fmt % line)
        self.buf.append(fmt % '')

    def handle_start_img(self, tag, attrs):
        self.buf.append('<cyan><u>')
        if 'alt' in attrs:
            self.buf.append('%s (%s)' % (attrs['alt'], attrs['src']))
        else:
            self.buf.append(attrs['src'])

    def handle_end_img(self, tag, attrs):
        self.buf.append('</u></cyan>')

    def getvalue(self):
        return ''.join(self.buf)

htmlconv = HTMLConv()


def html2vtml(vtmarkup):
    """ Convert hypertext markup into vt markup.
    The output can be given to `vtmlrender` for converstion to VT100
    sequences. """
    try:
        htmlconv.feed(vtmarkup)
        htmlconv.close()
        return htmlconv.getvalue()
    finally:
        htmlconv.reset()


def htmlrender(htmarkup, **kwargs):
    """ Process HTML into printable VTML. """
    return vtml.vtmlrender(html2vtml(htmarkup), **kwargs)


def htmlprint(*values, plain=None, **options):
    """ Convert HTML to VTML and then print it.
    Follows same semantics as vtmlprint. """
    print(*[htmlrender(x, plain=plain) for x in values], **options)



"""
Translate Markdown to VTML (via HTML) so it can be printed to a terminal or
saved to a text file in human readable format.
"""

import markdown2
from . import html

_md = markdown2.Markdown(extras=['fenced-code-blocks'])


def mdconvert(markdown):
    html = _md.convert(markdown)
    return html[:-1]


def mdrender(markdown, **kwargs):
    """ Process HTML into printable VTML. """
    return html.htmlrender(mdconvert(markdown), **kwargs)


def mdprint(*values, plain=None, **options):
    """ Convert HTML to VTML and then print it.
    Follows same semantics as vtmlprint. """
    print(*[mdrender(x, plain=plain) for x in values], **options)



"""
Data related tools/utilities.
"""

import collections
import collections.abc
import functools
import time

hone_cache_finders = {}
TTLCacheInfo = collections.namedtuple("TTLCacheInfo", ('hits', 'misses',
                                      'maxsize', 'currsize', 'maxage'))
HoneCacheInfo = collections.namedtuple('HoneCacheInfo', TTLCacheInfo._fields +
                                       ('partials', 'findger'))
TTLMappingEntry = collections.namedtuple("TTLMappingEntry", 'ctime, ref')


class TTLMapping(collections.abc.MutableMapping):
    """ Sized and aged mapping.  The time-to-live is fixed for the entire
    cache to simplify the garbage collector. """

    def __init__(self, maxsize=None, ttl=None, timer=time.monotonic,
                 container=collections.OrderedDict):
        self.data = container()
        self.maxsize = maxsize
        self.ttl = ttl
        self.timer = timer
        super().__init__()

    def is_expired(self, entry, time=None):
        time = self.timer() if time is None else time
        return time - entry.ctime > self.ttl

    def __delitem__(self, key):
        del self.data[key]

    def __iter__(self):
        """ Return valid keys as of each look into the iterator.  If the
        iterator is not used right away the keys returned will be those that
        are valid at view time and not when the iterator was created. """
        if self.ttl is None:
            return iter(self.data)
        else:
            return (k for k, e in self.data.items() if not self.is_expired(e))

    def __len__(self):
        self.gc()
        return len(self.data)

    def __setitem__(self, key, value):
        """ Wrap an item with a timestamp and if replacing an existing key
        move it to the end. """
        move = key in self.data
        time = self.timer() if self.ttl is not None else None
        entry = TTLMappingEntry(time, value)
        self.data[key] = entry
        if move:
            self.data.move_to_end(key)
        self.gc()

    def __getitem__(self, key):
        """ Unwrap and check the vitality of the entry. """
        entry = self.data[key]
        if self.ttl is not None and self.is_expired(entry):
            del self.data[key]
            raise KeyError(key)
        return entry.ref

    def __contains__(self, key):
        if self.ttl is None:
            return key in self.data
        try:
            return not self.is_expired(self.data[key])
        except KeyError:
            return False

    def gc(self):
        """ Garbage collect overflow and/or aged entries. """
        manifest = []
        overlimit = len(self.data) - self.maxsize \
            if self.maxsize is not None else 0
        now = self.ttl is not None and self.timer()
        for key, entry in self.data.items():
            if overlimit > 0 or (now and self.is_expired(entry, time=now)):
                overlimit -= 1
                manifest.append(key)
            else:
                break
        for x in manifest:
            del self.data[x]

    def clear(self):
        self.data.clear()


def hone_cache(maxsize=128, maxage=None, refineby='startswith',
               store_partials=False):
    """ A caching decorator that follows after the style of lru_cache. Calls
    that are sharper than previous requests are returned from the cache after
    honing in on the requested results using the `refineby` technique.

    The `refineby` technique can be `startswith`, `container` or a user
    defined function used to provide a subset of results.

    Eg.
        @hone_cache()
        def completer(prefix):
            print("MISS")
            return set(x for x in {'foo', 'foobar', 'baz'}
                       if x.startswith(prefix))
        completer('f')    # -> {'foo', 'foobar'}
        MISS
        completer('fo')   # -> {'foo', 'foobar'}
        completer('foob') # -> {'foobar'}
        completer('')     # -> {'foo', 'foobar', 'baz'}
        MISS
        completer('ba')   # -> {'baz'}

    If while trying to hone a cache hit there is an exception the wrapped
    function will be called as if it was a full miss.

    The `maxage` argument is an option time-to-live in seconds for each cache
    result.  Any cache entries over the maxage are lazily replaced. """
    if not callable(refineby):
        finder = hone_cache_finders[refineby]
    else:
        finder = refineby

    def decorator(inner_func):
        wrapper = make_hone_cache_wrapper(inner_func, maxsize, maxage, finder,
                                          store_partials)
        return functools.update_wrapper(wrapper, inner_func)
    return decorator


def hone_cache_startswith_finder(radix, partial_radix, partial):
    return type(partial)(x for x in partial if x.startswith(radix))

hone_cache_finders['startswith'] = hone_cache_startswith_finder


def hone_cache_container_finder(radix, partial_radix, partial):
    path_offt = radix[len(partial_radix):]
    for x in path_offt:
        partial = partial[x]
    return partial

hone_cache_finders['container'] = hone_cache_container_finder


def make_hone_cache_wrapper(inner_func, maxsize, maxage, finder,
                            store_partials):
    """ Keeps a cache of requests we've already made and use that for
    generating results if possible.  If the user asked for a root prior
    to this call we can use it to skip a new lookup using `finder`.  A
    top-level lookup will effectively serves as a global cache. """

    hits = misses = partials = 0
    cache = TTLMapping(maxsize, maxage)

    def wrapper(*args):
        nonlocal hits, misses, partials
        radix = args[-1]
        # Attempt fast cache hit first.
        try:
            r = cache[radix]
        except KeyError:
            pass
        else:
            hits += 1
            return r
        for i in range(len(radix) - 1, -1, -1):
            partial_radix = radix[:i]
            try:
                partial = cache[partial_radix]
            except KeyError:
                continue
            try:
                r = finder(radix, partial_radix, partial)
            except:
                break  # Treat any exception as a miss.
            partials += 1
            if store_partials:
                cache[radix] = r
            return r
        misses += 1
        cache[radix] = r = inner_func(*args)
        return r

    def cache_info():
        """ Emulate lru_cache so this is a low touch replacement. """
        return HoneCacheInfo(hits, misses, maxsize, len(cache), maxage,
                             partials, finder)

    def cache_clear():
        """ Clear cache and stats. """
        nonlocal hits, misses, partials
        hits = misses = partials = 0
        cache.clear()

    wrapper.cache_info = cache_info
    wrapper.cache_clear = cache_clear

    return functools.update_wrapper(wrapper, inner_func)


def ttl_cache(maxage, maxsize=128):
    """ A time-to-live caching decorator that follows after the style of
    lru_cache.  The `maxage` argument is time-to-live in seconds for each
    cache result.  Any cache entries over the maxage are lazily replaced. """

    def decorator(inner_func):
        wrapper = make_ttl_cache_wrapper(inner_func, maxage, maxsize)
        return functools.update_wrapper(wrapper, inner_func)
    return decorator


def make_ttl_cache_wrapper(inner_func, maxage, maxsize, typed=False):
    """ Use the function signature as a key for a ttl mapping.  Any misses
    will defer to the wrapped function and its result is stored for future
    calls. """

    hits = misses = 0
    cache = TTLMapping(maxsize, maxage)

    def wrapper(*args, **kwargs):
        nonlocal hits, misses
        key = functools._make_key(args, kwargs, typed)
        try:
            result = cache[key]
        except KeyError:
            misses += 1
            result = cache[key] = inner_func(*args, **kwargs)
        else:
            hits += 1
        return result

    def cache_info():
        """ Emulate lru_cache so this is a low touch replacement. """
        return TTLCacheInfo(hits, misses, maxsize, len(cache), maxage)

    def cache_clear():
        """ Clear cache and stats. """
        nonlocal hits, misses
        hits = misses = 0
        cache.clear()

    wrapper.cache_info = cache_info
    wrapper.cache_clear = cache_clear

    return functools.update_wrapper(wrapper, inner_func)



"""
Functions for displaying content with screen aware layout.
"""

import enum
import functools
import html.parser
import re
import sys

TAGS = {
    'b': 1,
    'dim': 2,
    'i': 3,
    'u': 4,
    'blink': 5,
    'reverse': 7,
    'black': 30,
    'red': 31,
    'green': 32,
    'yellow': 33,
    'blue': 34,
    'magenta': 35,
    'cyan': 36,
    'white': 37,
    'bgblack': 40,
    'bgred': 41,
    'bggreen': 42,
    'bgyellow': 43,
    'bgblue': 44,
    'bgmagenta': 45,
    'bgcyan': 46,
    'bgwhite': 47,
}


def beststr(*strings):
    """ Test if the output device can handle the desired strings. The options
    should be sorted by preference. Eg. beststr(unicode, ascii). """
    for x in strings:
        try:
            x.encode(sys.stdout.encoding)
        except UnicodeEncodeError:
            pass
        else:
            return x
    raise ValueError('No valid strings found')

# Regex for handling line wrapping...
#  * Splits by spaces, newlines and hypens.
#  * Hypens are kept on leftmost word.
#  * Newlines are not grouped with other whitespace.
#  * Other whitespace is grouped.
_textwrap_word_break = re.compile('(\n|[ \t\f\v\r]+|[^\s]+?-+)')
_whitespace = re.compile('[ \t\f\v\r]+')


def is_whitespace(value):
    return not not _whitespace.match(value)


def _add_slice(seq, slc):
    """ Our textwrap routine deals in slices.  This function will concat
    contiguous slices as an optimization so lookup performance is faster.
    It expects a sequence (probably a list) to add slice to or will extend
    the last slice of the sequence if it ends where the new slice begins. """
    if seq and seq[-1].stop == slc.start:
        seq[-1] = slice(seq[-1].start, slc.stop)
    else:
        seq.append(slc)


def _textwrap_slices(text, width, strip_leading_indent=False):
    """ Nearly identical to textwrap.wrap except this routine is a tad bit
    safer in its algo that textwrap.  I ran into some issues with textwrap
    output that make it unusable to this usecase as a baseline text wrapper.
    Further this utility returns slices instead of strings.  So the slices
    can be used to extract your lines manually. """
    if not isinstance(text, str):
        raise TypeError("Expected `str` type")
    chunks = (x for x in _textwrap_word_break.split(text) if x)
    remaining = width
    buf = []
    lines = [buf]
    whitespace = []
    whitespace_len = 0
    pos = 0
    try:
        chunk = next(chunks)
    except StopIteration:
        chunk = ''
    if not strip_leading_indent and is_whitespace(chunk):
        # Add leading indent for first line, but only up to one lines worth.
        chunk_len = len(chunk)
        if chunk_len >= width:
            _add_slice(buf, slice(0, width))
            buf = []
            lines.append(buf)
        else:
            _add_slice(buf, slice(0, chunk_len))
            remaining -= chunk_len
        pos = chunk_len
        try:
            chunk = next(chunks)
        except StopIteration:
            chunk = ''
    while True:
        avail_len = remaining - whitespace_len
        chunk_len = len(chunk)
        if chunk == '\n':
            buf = []
            lines.append(buf)
            whitespace = []
            whitespace_len = 0
            remaining = width
        elif is_whitespace(chunk):
            if buf:
                _add_slice(whitespace, slice(pos, pos + chunk_len))
                whitespace_len += chunk_len
        elif len(chunk) > avail_len:
            if not buf:
                # Must hard split the chunk.
                for x in whitespace:
                    _add_slice(buf, x)
                _add_slice(buf, slice(pos, pos + avail_len))
                chunk = chunk[avail_len:]
                pos += avail_len
            # Bump to next line without fetching the next chunk.
            buf = []
            lines.append(buf)
            whitespace = []
            whitespace_len = 0
            remaining = width
            continue
        else:
            if buf:
                remaining -= whitespace_len
                for x in whitespace:
                    _add_slice(buf, x)
            whitespace = []
            whitespace_len = 0
            _add_slice(buf, slice(pos, pos + chunk_len))
            remaining -= chunk_len
        pos += chunk_len
        try:
            chunk = next(chunks)
        except StopIteration:
            break
    return lines


class VTMLParser(html.parser.HTMLParser):
    """ Add some SGML style tag support for a few VT100 operations. """

    escape = ('&',)
    escape_setinal_base = 0xe2c6  # protected by PAU

    def __init__(self, *args, **kwargs):
        escape_tuples = enumerate(self.escape, self.escape_setinal_base)
        self.escape_map = tuple((chr(i), x) for i, x in escape_tuples)
        super().__init__(*args, **kwargs)

    def feed(self, data):
        """ We need to prevent insertion of some special HTML characters
        (entity refs and maybe more).  It's hard to prevent the state machine
        from blowing up with them in the data, so we temporarily replacing
        them with a reserved unicode value based on the PAU private space. """
        assert not self.closed
        for mark, special in self.escape_map:
            assert mark not in data
            data = data.replace(special, mark)
        super().feed(data)

    def reset(self):
        self.closed = False
        self.vbuf = VTMLBuffer()
        self.open_tags = []
        super().reset()

    def handle_starttag(self, tag, attrs):
        if tag not in TAGS:
            return self.handle_data(self.get_starttag_text())
        self.open_tags.append(tag)

    def handle_endtag(self, tag):
        if tag not in TAGS:
            return self.handle_data("</%s>" % tag)
        if self.open_tags[-1] != tag:
            raise SyntaxError("Bad close tag: %s; Expected: %s" % (tag,
                              self.open_tags[-1]))
        del self.open_tags[-1]
        self.vbuf.append_reset()

    def handle_data(self, data):
        if self.open_tags:
            for tag in self.open_tags:
                self.vbuf.append_tag(tag)
        for mark, special in self.escape_map:
            data = data.replace(mark, special)
        self.vbuf.append_str(data)

    def handle_entityref(self, name):
        raise RuntimeError("programmer error")

    def handle_charref(self, name):
        raise RuntimeError("programmer error")

    def unknown_decl(self, name):
        raise RuntimeError("programmer error")

    def handle_comment(self, data):
        raise RuntimeError("programmer error")

    def close(self):
        assert not self.closed
        super().close()
        if self.open_tags:
            self.vbuf.append_reset()
        self.closed = True

    def getvalue(self):
        assert self.closed
        return self.vbuf


@functools.total_ordering
class VTMLBuffer(object):
    """ A str-like object that has an adjusted length to compensate for
    nonvisual vt100 opcodes which do not occupy space in the output. """

    ops = enum.Enum('ops', 'reset tag str')
    _reset_opcode = '\033[0m'

    @classmethod
    def new(cls, *args, **kwargs):
        """ Subclass friendly factory for getting a new buffer. """
        return cls(*args, **kwargs)

    @classmethod
    def from_buffers(cls, buffers):
        """ Join all argument buffers into a new unified buffer object. """
        new = cls()
        for obj in buffers:
            new.extend(obj)
        return new

    def __init__(self, value=None):
        self._values = []
        if value is not None:
            if isinstance(value, str):
                self.append_str(value)
            elif isinstance(value, VTMLBuffer):
                self.extend(value)
            else:
                raise TypeError("Init value must be `str` or `VTMLBuffer`")

    def __len__(self):
        return len(self.text())

    def __str__(self):
        buf = []
        for op, val in self._values:
            if op == self.ops.str:
                buf.append(val)
            elif op == self.ops.tag:
                buf.append('\033[%dm' % TAGS[val])
            elif op == self.ops.reset:
                buf.append(self._reset_opcode)
            else:
                raise ValueError("invalid op: %r" % (op,))
        return ''.join(buf)

    def __repr__(self):
        return repr(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __contains__(self, other):
        return other in self.text()

    def __format__(self, fmt):
        """ Add support for re-embedded VTML via the `vtml` specifier in a
        `str.format` argument. E.g.
            >>> words = shellish.vtmlrender('<u>underlined words')
            >>> '<b>Make bold: {:vtml}, thanks</b>'.format(words)
            '<b>Make bold: <u>underlined words</u>, thanks</b>'
        """
        if fmt == 'vtml':
            buf = []
            tag_stack = []
            for op, val in self._values:
                if op == self.ops.str:
                    buf.append(val)
                elif op == self.ops.tag:
                    tag_stack.append(val)
                    buf.append('<%s>' % val)
                elif op == self.ops.reset:
                    if tag_stack:
                        for x in reversed(tag_stack):
                            buf.append('</%s>' % x)
                        del tag_stack[:]
                else:
                    raise ValueError("invalid op: %r" % (op,))
            return ''.join(buf)
        else:
            return str(self)

    def __add__(self, other):
        if isinstance(other, str):
            other = self.new(other)
        elif not isinstance(other, VTMLBuffer):
            raise TypeError("Invalid concatenation type: %s" % type(other))
        new = self.copy()
        new.extend(other)
        return new

    def __iadd__(self, other):
        if isinstance(other, str):
            self.append_str(other)
        elif isinstance(other, VTMLBuffer):
            self.extend(other)
        else:
            raise TypeError("Invalid concatenation type: %s" % type(other))
        return self

    def __radd__(self, other):
        if isinstance(other, str):
            return other + str(self)
        elif isinstance(other, VTMLBuffer):
            return other + self
        else:
            raise TypeError("Invalid concatenation type: %s" % type(other))

    def __mul__(self, factor):
        if not isinstance(factor, int):
            raise TypeError('Expected `int` type factor')
        new = self.copy()
        new._values *= factor
        return new

    __rmul__ = __mul__

    def __imul__(self, factor):
        if not isinstance(factor, int):
            raise TypeError('Expected `int` type factor')
        self._values *= factor
        return self

    def __getitem__(self, key):
        """ Support for slicing and indexing.  Results are always a new
        VTMLBuffer copy. """
        visual_len = len(self)
        if isinstance(key, slice):
            if key.step is not None:
                raise TypeError("`step` is not supported")
            start = key.start or 0
            if start < 0:
                start = visual_len + start
            stop = key.stop if key.stop is not None else visual_len
            if stop < 0:
                stop = visual_len + stop
        else:
            start = key
            if start < 0:
                start = visual_len + start
            if not 0 <= start < visual_len:
                raise IndexError('Index out of range')
            stop = start + 1
        remaining = max(stop - start, 0)
        new = self.new()
        op_active = False
        pos = 0
        for op, val in self._values:
            if not remaining:
                break
            elif op == self.ops.tag:
                new.append_tag(val)
                op_active = True
            elif op == self.ops.reset:
                new.append_reset()
                op_active = False
            elif op == self.ops.str:
                snip = val[start - pos:] if start > pos else val
                pos += len(val)
                if snip:
                    fragment = snip[:remaining]
                    remaining -= len(fragment)
                    new.append_str(fragment)
            else:
                raise RuntimeError("Invalid op type")
        if op_active:
            new.append_reset()
        return new

    def copy(self):
        return self.new(self)

    def append_reset(self):
        self._values.append((self.ops.reset, None))

    def append_tag(self, tag):
        self._values.append((self.ops.tag, tag))

    def append_str(self, value):
        self._values.append((self.ops.str, value))

    def extend(self, buf):
        if not isinstance(buf, VTMLBuffer):
            raise TypeError("Expected `VTMLBuffer`")
        self._values.extend(buf._values)

    def _promiscuous_extend(self, buf, other):
        """ Extend that supports `VTMLBuffer` and `str`. """
        try:
            buf.extend(other)
        except TypeError:
            if isinstance(other, str):
                buf.append_str(other)
            else:
                raise TypeError("Expected `VTMLBuffer` or `str`. Got `%s`" %
                                type(other))

    def join(self, buffers):
        """ Same interface as b''.join and ''.join. Supports upconversion of
        `str` types too. """
        output = self.new()
        bufiter = iter(buffers)
        try:
            self._promiscuous_extend(output, next(bufiter))
        except StopIteration:
            return output
        for buf in bufiter:
            output.extend(self)
            self._promiscuous_extend(output, buf)
        return output

    def text(self):
        """ Return just the text content of this string without opcodes. """
        return ''.join(val for op, val in self._values if op == self.ops.str)

    def plain(self):
        """ Similar to `text` but returns valid VTMLBuffer instance. """
        new = self.new()
        for op, val in self._values:
            if op == self.ops.str:
                new.append_str(val)
        return new

    def clip(self, length, cliptext=''):
        """ Clip text for lines exceeding a particular length.  Newlines and
        trailing are also removed. """
        if length < 0:
            raise ValueError("Negative clip invalid")
        cliplen = len(cliptext)
        if length < cliplen:
            raise ValueError("Clip length too small: %d < %d" % (length,
                             cliplen))
        text = self.text()
        first = text.splitlines()[0] if text else text
        stripped = first.rstrip()
        clipping = len(first) != len(text) or len(stripped) > length
        adj_length = min(len(stripped), length - (cliplen if clipping else 0))
        new = self[:adj_length]
        if clipping and cliptext:
            new.append_str(cliptext)
        return new

    def wrap(self, width, **options):
        """ Text wrapping similar to textwrap.wrap but protects vt escape
        sequences by returning a list of VTMLBuffer objects. """
        if width <= 0:
            raise ValueError("Invalid wrap width: %d" % width)
        slices = _textwrap_slices(self.text(), width, **options)
        return [self.from_buffers(self[s] for s in line_slices)
                for line_slices in slices]

    def ljust(self, width, fillchar=' '):
        new = self.copy()
        if width > len(self):
            new.append_str(fillchar * (width - len(self)))
        return new

    def rjust(self, width, fillchar=' '):
        new = self.new()
        if width > len(self):
            new.append_str(fillchar * (width - len(self)))
        new.extend(self)
        return new

    def center(self, width, fillchar=' '):
        """ Center strings so uneven padding always favors trailing pad.  When
        centering clumps of text this produces better results than str.center
        which alternates which side uneven padding occurs on. """
        if width > len(self):
            padlen = width - len(self)
            leftlen = padlen // 2
            rightlen = padlen - leftlen
            new = self.new()
            new.append_str(fillchar * leftlen)
            new.extend(self)
            new.append_str(fillchar * rightlen)
            return new
        else:
            return self.copy()

    def rstrip(self):
        """ Removing trailing whitespace. """
        removals = []
        i = len(self._values)
        for op, val in reversed(self._values):
            i -= 1
            if op == self.ops.str:
                if is_whitespace(val):
                    removals.append(i)
                else:
                    break
        copy = self.copy()
        for i in removals:
            del copy._values[i]
        return copy

    def startswith(self, other):
        return self.text().startswith(other)

    def endswith(self, other):
        return self.text().endswith(other)

    def split(self, sep=' ', maxsplit=None):
        if not sep:
            raise ValueError("empty separator")
        splits = []
        pos = 0
        text = self.text()
        sep_len = len(sep)
        while maxsplit is None or len(splits) < maxsplit:
            try:
                end = text.index(sep, pos)
            except ValueError:
                break
            splits.append((pos, end))
            pos = end + sep_len
        splits.append((pos, None))
        return [self[start:end] for start, end in splits]
        if splits:
            return [self[start:end] for start, end in splits]
        else:
            return [self.copy()]


def vtmlrender(vtmarkup, plain=None, strict=False, vtmlparser=VTMLParser()):
    """ Look for vt100 markup and render vt opcodes into a VTMLBuffer. """
    if isinstance(vtmarkup, VTMLBuffer):
        return vtmarkup.plain() if plain else vtmarkup
    try:
        vtmlparser.feed(vtmarkup)
        vtmlparser.close()
    except:
        if strict:
            raise
        buf = VTMLBuffer()
        buf.append_str(str(vtmarkup))
        return buf
    else:
        buf = vtmlparser.getvalue()
        return buf.plain() if plain else buf
    finally:
        vtmlparser.reset()


def vtmlprint(*values, plain=None, strict=None, **options):
    """ Follow normal print() signature but look for vt100 codes for richer
    output. """
    print(*[vtmlrender(x, plain=plain, strict=strict) for x in values],
          **options)



"""
Alternate form of getting help.
"""

import difflib
import sys
from .. import command
from ... import rendering


class Help(command.Command):
    """ Show help for a command. """

    name = 'help'
    use_pager = True

    def setup_args(self, parser):
        self.add_argument('command', nargs='*', complete=self.command_choices,
                          help='Command name to show help of.')

    def command_choices(self, prefix, args):
        if args.command:
            cmd = self.find_command(args.command[:-1 if prefix else None])
        else:
            cmd = self.find_root()
        return frozenset(x for x in cmd.subcommands if x.startswith(prefix))

    def stderr(self, *args, **kwargs):
        return rendering.vtmlprint(*args, file=sys.stderr, **kwargs)

    def find_command(self, path):
        cmdptr = self.find_root()
        for i, cmd in enumerate(path):
            try:
                cmdptr = cmdptr.subcommands[cmd]
            except KeyError:
                raise LookupError(cmdptr, i)
        return cmdptr

    def run(self, args):
        if not args.command:
            self.find_root().argparser.print_help()
            return
        try:
            cmd = self.find_command(args.command)
        except LookupError as e:
            good_cmd, fail_idx = e.args
            good = ' '.join(args.command[:fail_idx]) + ' ' if fail_idx else ''
            bad = ' '.join(args.command[fail_idx:])
            self.stderr('<red>Command not found:</red> %s<b><red>%s' % (good,
                        bad))
            for x in difflib.get_close_matches(args.command[fail_idx],
                                               good_cmd.subcommands):
                self.stderr('    Did you mean: %s<b>%s</b>?' % (good, x))
            raise SystemExit(1)
        else:
            cmd.argparser.print_help()



"""
Tab completion handling.
"""

import argparse
import os


class ActionCompleter(object):
    """ Stateful behavior for tab completion.  Calling this instance returns
    valid choices for the action. """

    sentinel = ' '

    def __init__(self, action):
        self.action = action
        self.key = None
        self.choices = None
        self.consumed = 0
        self.calling_command = None  # Only set during completion.
        self.subparsers = None
        if action.choices and hasattr(action.choices, 'items'):
            self.subparsers = action.choices.copy()
        if action.option_strings:
            # Only include the longest option string under the assumption
            # that this most accurately describes the argument.
            self.key = max(action.option_strings, key=len)
        if action.choices:
            self.completer = self.choice_complete
            self.choices = action.choices
        elif getattr(action, 'complete', None):
            self.completer = self.proxy_complete(action.complete)
        elif isinstance(action.type, argparse.FileType):
            self.completer = self.file_complete
        else:
            self.completer = self.hint_complete
        self.parse_nargs(action.nargs)

    def __str__(self):
        return '<%s key:%s action:(%s)>' % (type(self).__name__, self.key,
                                            self.about_action())

    __repr__ = __str__

    def __call__(self, command, prefix, args):
        args = self.silent_parse_args(command, args)
        self.calling_command = command
        try:
            return self.completer(prefix, args)
        finally:
            self.calling_command = None

    def silent_parse_args(self, command, args):
        """ Silently attempt to parse args.  If there is a failure then we
        ignore the effects.  Using an in-place namespace object ensures we
        capture as many of the valid arguments as possible when the argparse
        system would otherwise throw away the results. """
        args_ns = argparse.Namespace()
        stderr_save = argparse._sys.stderr
        stdout_save = argparse._sys.stdout
        argparse._sys.stderr = os.devnull
        argparse._sys.stdout = os.devnull
        try:
            command.argparser.parse_known_args(args, args_ns)
        except BaseException:
            pass
        finally:
            argparse._sys.stderr = stderr_save
            argparse._sys.stdout = stdout_save
        return args_ns

    def parse_nargs(self, nargs):
        """ Nargs is essentially a multi-type encoding.  We have to parse it
        to understand how many values this action may consume. """
        self.max_args = self.min_args = 0
        if nargs is None:
            self.max_args = self.min_args = 1
        elif nargs == argparse.OPTIONAL:
            self.max_args = 1
        elif nargs == argparse.ZERO_OR_MORE:
            self.max_args = None
        elif nargs in (argparse.ONE_OR_MORE, argparse.REMAINDER):
            self.min_args = 1
            self.max_args = None
        elif nargs != argparse.PARSER:
            self.max_args = self.min_args = nargs

    def consume(self, args):
        """ Consume the arguments we support.  The args are modified inline.
        The return value is the number of args eaten. """
        consumable = args[:self.max_args]
        self.consumed = len(consumable)
        del args[:self.consumed]
        return self.consumed

    @property
    def full(self):
        """ Can this action take more arguments? """
        if self.max_args is None:
            return False
        return self.consumed >= self.max_args

    @property
    def reached_min(self):
        """ Have we consumed the minimum number of args. """
        return self.consumed >= self.min_args

    def about_action(self):
        """ Simple string describing the action. """
        name = self.action.metavar or self.action.dest
        type_name = self.action.type.__name__ if self.action.type else ''
        if self.action.help or type_name:
            extra = ' (%s)' % (self.action.help or 'type: %s' % type_name)
        else:
            extra = ''
        return name + extra

    def choice_complete(self, prefix, args):
        return frozenset(x for x in self.choices if x.startswith(prefix))

    def file_complete(self, prefix, args):
        """ Look in the local filesystem for valid file choices. """
        path = os.path.expanduser(prefix)
        dirname, name = os.path.split(path)
        if not dirname:
            dirname = '.'
        try:
            dirs = os.listdir(dirname)
        except FileNotFoundError:
            return frozenset()
        choices = []
        session = self.calling_command.session
        for f in dirs:
            try:
                if (not name or f.startswith(name)) and \
                   not f.startswith('.'):
                    choices.append(f)
            except PermissionError:
                pass
        prevent_pad = session.pad_completion and len(choices) == 1 and \
            os.path.isdir(choices[0])
        names = [os.path.join(dirname, x) for x in choices]
        if prevent_pad:
            names.append(names[0] + '/')
        return frozenset(names)

    def hint_complete(self, prefix, args):
        """ For arguments that don't have complete functions or .choices we
        can only hint about the argument details. The results are designed to
        not self-expand (ie, len(choices) > 1). """
        return frozenset(('[%s]' % self.about_action(), self.sentinel))

    def proxy_complete(self, func):
        """ Pass completion work to foreign function. """
        return lambda *args, **kwargs: frozenset(func(*args, **kwargs))



"""
Visual progress indicators.
"""

import math
import shutil
import sys
from .. import rendering


class ProgressBar(object):
    """ Progress bar class for more advanced control apps.  Most likely this
    should just be used by subclasses or by the generator functions. """

    partials = rendering.beststr('▏▎▎▍▍▌▌▋▋▊▊▉▉█', '#')
    fill = rendering.beststr('▯', '-')

    def __init__(self, min=0, max=1, width=None, file=None, prefix=' ',
                 suffix=' ', bar_style='', fill_style='', show_percent=True,
                 clear=False):
        if width is None:
            width = shutil.get_terminal_size()[0]
        if file is None:
            self.file = sys.stderr
        self.clear_on_exit = clear
        self.min_value = min
        self.max_value = max
        self.width = width
        self.bar_width = width - len(prefix) - len(suffix)
        if show_percent:
            self.bar_width -= 5
        self.prefix = prefix
        self.suffix = suffix
        self.bar_style = bar_style
        self.fill_style = fill_style
        self.show_percent = show_percent

    def __enter__(self):
        self.hide_cursor()
        return self

    def __exit__(self, *exc):
        if self.clear_on_exit:
            self.clear()
        else:
            print(file=self.file)
        self.restore_cursor()

    def draw(self):
        pct = (self.value - self.min_value) / self.max_value
        chars = pct * self.bar_width
        partial = chars % 1
        partial_idx = math.floor(partial * len(self.partials))
        wholechars = math.floor(chars)
        bar = (self.partials[-1] * wholechars)
        if len(bar) < self.bar_width:
            bar += self.partials[partial_idx]
        fill = self.fill * (self.bar_width - len(bar))
        if self.show_percent:
            pct = '%4.0f%%' % (pct * 100)
        rendering.vtmlprint('\r', self.prefix, self.bar_style + bar,
                            self.fill_style + fill, pct, self.suffix, sep='',
                            end='', file=self.file, flush=True)

    def set(self, value):
        self.value = value
        self.draw()

    def clear(self):
        rendering.vtmlprint('\r', ' ' * self.width, file=self.file)

    def hide_cursor(self):
        print('\033[?25l', file=self.file)

    def restore_cursor(self):
        print('\033[?25h', file=self.file)


def progressbar(stream, prefix='Loading: ', width=0.5, **options):
    """ Generator filter to print a progress bar. """
    size = len(stream)
    if not size:
        return stream
    if 'width' not in options:
        if width <= 1:
            width = round(shutil.get_terminal_size()[0] * width)
        options['width'] = width
    with ProgressBar(max=size, prefix=prefix, **options) as b:
        b.set(0)
        for i, x in enumerate(stream, 1):
            yield x
            b.set(i)



"""
A command class that is built from a normal function.
"""

import argparse
import inspect
import os.path
import sys
from . import command


class AutoCommand(command.Command):
    """ Auto command ABC.  This command wraps a generic function and tries
    to map the function signature to a parser configuration.  Use the
    @autocommand decorator to use it. """

    falsy = {
        'false',
        'none',
        'null',
        '0',
        'no',
        'off',
        'disable'
    }

    def __init__(self, *args, func=None, **kwargs):
        self.func = func
        super().__init__(*args, **kwargs)

    def run(self, args):
        """ Convert the unordered args into function arguments. """
        args = vars(args)
        positionals = []
        keywords = {}
        for action in self.argparser._actions:
            if not hasattr(action, 'label'):
                continue
            if action.label == 'positional':
                positionals.append(args[action.dest])
            elif action.label == 'varargs':
                positionals.extend(args[action.dest])
            elif action.label == 'keyword':
                keywords[action.dest] = args[action.dest]
            elif action.label == 'varkwargs':
                kwpairs = iter(args[action.dest] or [])
                for key in kwpairs:
                    try:
                        key, value = key.split('=', 1)
                    except ValueError:
                        value = next(kwpairs)
                        key = key.strip('-')
                    keywords[key] = value
        return self.func(*positionals, **keywords)

    def setup_args(self, default_parser):
        sig = inspect.signature(self.func)
        got_keywords = False
        for name, param in sig.parameters.items():
            label = ''
            options = {}
            help = None
            parser = default_parser
            if param.annotation is not sig.empty:
                options['type'] = param.annotation
            elif param.default not in (sig.empty, None):
                options['type'] = type(param.default)
            if param.kind in (param.POSITIONAL_OR_KEYWORD,
                              param.KEYWORD_ONLY):
                if param.kind == param.KEYWORD_ONLY or \
                   param.default is not sig.empty:
                    name = '--%s' % name.replace('_', '-')
                    label = 'keyword'
                    got_keywords = True
                else:
                    label = 'positional'
            elif param.kind == param.VAR_POSITIONAL:
                if got_keywords:
                    # Logically impossible to handle this since keyword
                    # arguments are always expressed as key/value pairs and
                    # this language feature is based on giving a keyword
                    # argument by just its position.
                    raise ValueError("Unsupported function signature: %s" %
                                     sig)
                help = "variable positional args"
                options['nargs'] = '*'
                name = '*%s' % name
                label = 'varargs'
            elif param.kind == param.VAR_KEYWORD:
                options['nargs'] = argparse.REMAINDER
                name = '--%s' % name.replace('_', '-')
                label = 'varkwargs'
                help = 'variable key/value args [[--key value] || ' \
                       '[key=value] ...]'
            if param.default not in (sig.empty, None):
                options['default'] = param.default
            if help:
                options['help'] = help
            if options.get('type'):
                if options['type'] is bool:
                    if label == 'keyword':
                        options['action'] = 'store_%s' % \
                            str(not param.default).lower()
                        del options['type']
                    else:
                        options['type'] = lambda x: x.lower() not in self.falsy
            action = parser.add_argument(name, **options)
            action.label = label


def autocommand(func):
    """ A simplified decorator for making a single function a Command
    instance.  In the future this will leverage PEP0484 to do really smart
    function parsing and conversion to argparse actions. """
    if func.__module__ == '__main__':
        name = os.path.basename(sys.argv[0])
    else:
        name = func.__name__
    title, desc = command.parse_docstring(func)
    if not title:
        title = 'Auto command for: %s' % name
    if not desc:
        # Prevent Command from using docstring of AutoCommand
        desc = ' '
    return AutoCommand(title=title, desc=desc, name=name, func=func)



"""
Session context for commands.  This provides singular state mgmt for a tree
of commands as well as the implementation for interactive mode.
"""

import ast
import configparser
import contextlib
import os.path
import pdb
import readline
import shutil
import sys
import traceback
from . import eventing, rendering


def _vprinterr(*args, **kwargs):
    return rendering.vtmlprint(*args, file=sys.stderr, **kwargs)


class SessionExit(BaseException):
    pass


class Session(eventing.Eventer):
    """ The session manager for a tree of commands. """

    var_dir = os.path.expanduser('~')
    command_error_verbosity = 'traceback'
    default_prompt_format = '[{name}] $'
    intro = 'Type "help" or "?" to list commands and "exit" to quit.'
    completer_delim_includes = frozenset()
    completer_delim_excludes = frozenset('-+@:/~*')
    pad_completion = True
    allow_pager = bool(os.environ.get('PAGER', True))

    @property
    def config_file(self):
        return '.%s_config' % self.name

    @property
    def history_file(self):
        return '.%s_history' % self.name

    def __init__(self, root_command, name=None):
        self.root_command = root_command
        self.name = name or root_command.name
        self.config = self.load_config()
        self.add_events(['precmd', 'postcmd'])
        raw_prompt = self.config['ui']['prompt_format']
        self.prompt_format = ast.literal_eval("'%s '" % raw_prompt)
        super().__init__()

    def default_config(self):
        pager = os.environ.get('PAGER')
        if not pager:
            pager = "less -i -X -R -F -P's{desc}, line %l" \
                    "(press h for help or q to quit)'"
        return {
            "ui": {
                "prompt_format": self.default_prompt_format
            },
            "core": {
                "pager": pager
            }
        }

    def load_config(self):
        filename = os.path.join(self.var_dir, self.config_file)
        config = configparser.ConfigParser(interpolation=None)
        config.read_dict(self.default_config())
        config.read_dict(self.command_default_configs())
        config.read(filename)
        return config

    def load_history(self):
        filename = os.path.join(self.var_dir, self.history_file)
        try:
            readline.read_history_file(filename)
        except FileNotFoundError:
            pass
        return filename

    def exit(self):
        """ Can be called by commands to stop an interactive session. """
        raise SessionExit()

    def map_subcommands(self, func):
        """ Run `func` against all the subcommands attached to our root
        command. """

        def crawl(cmd):
            for sc in cmd.subcommands.values():
                yield from crawl(sc)
            yield cmd
        return map(func, crawl(self.root_command))

    def command_default_configs(self):
        getconfig = lambda cmd: (cmd.config_section(), cmd.default_config())
        return dict(self.map_subcommands(getconfig))

    def execute(self, command, args):
        """ Event firing and exception conversion around command execution.
        Common exceptions are run through our exception handler for
        pretty-printing or debugging and then converted to SystemExit so the
        interpretor will exit without further ado (or be caught if
        interactive). """
        self.fire_event('precmd', command, args)
        try:
            try:
                result = command.run_wrap(args)
            except BaseException as e:
                self.fire_event('postcmd', command, args, exc=e)
                raise e
            else:
                self.fire_event('postcmd', command, args, result=result)
                return result
        except BrokenPipeError as e:
            _vprinterr('<dim><red>...broken pipe...</red></dim>')
            raise SystemExit(1) from e
        except KeyboardInterrupt as e:
            _vprinterr('<dim><red>...interrupted...</red></dim>')
            raise SystemExit(1) from e
        except SystemExit as e:
            if e.args and not isinstance(e.args[0], int):
                _vprinterr("<red>%s</red>" % e)
                raise SystemExit(1) from e
            raise e
        except Exception as e:
            self.handle_command_error(command, args, e)
            raise SystemExit(1) from e

    def handle_command_error(self, command, args, exc):
        """ Depending on how the session is configured this will print
        information about an unhandled command exception or possibly jump
        to some other behavior like a debugger. """
        verbosity = self.command_error_verbosity
        if verbosity == 'traceback':
            self.pretty_print_exc(command, exc, show_traceback=True)
        elif verbosity == 'debug':
            pdb.set_trace()
        elif verbosity == 'raise':
            raise exc
        elif verbosity == 'pretty':
            self.pretty_print_exc(command, exc)
        else:
            raise ValueError('Unexpected exception_verbosity: %s' %
                             verbosity)

    def pretty_print_exc(self, command, exc, show_traceback=False):
        cmdname = command.prog or command.name
        if not show_traceback:
            _vprinterr("<red>Command '%s' error: %s - %s</red>" % (cmdname,
                       type(exc).__name__, exc))
        else:
            _vprinterr("<red>Command '%s' error, traceback...</red>" %
                       cmdname)
            rendering.print_exception(exc, file=sys.stderr)

    @property
    def prompt(self):
        raw = self.prompt_format.format(**self.prompt_info())
        return rendering.vtmlrender(raw)

    def prompt_info(self):
        """ Return a dictionary of items that can be substituted into the
        prompt_format by the subclass or its user if customized in the
        config file. """
        return {
            "name": self.name,
        }

    def cmd_split(self, line):
        """ Get the command associated with this input line. """
        cmd, *args = line.lstrip().split(' ', 1)
        return self.root_command.subcommands[cmd], ' '.join(args)

    def completer_hook(self, text, state):
        if state == 0:
            linebuf = readline.get_line_buffer()
            line = linebuf.lstrip()
            stripped = len(linebuf) - len(line)
            begin = readline.get_begidx() - stripped
            end = readline.get_endidx() - stripped
            if begin > 0:
                try:
                    cmd, args = self.cmd_split(line)
                except KeyError:
                    return None
                cfunc = cmd.complete
            else:
                cfunc = self.complete_names
            if self.pad_completion:
                pad = lambda x: x + ' ' if not x.endswith(' ') or \
                    x.endswith(r'\ ') else x
            else:
                pad = lambda x: x
            choices = self.complete_wrap(cfunc, text, line, begin, end)
            self.completer_cache = list(map(pad, choices))
        try:
            return self.completer_cache[state]
        except IndexError:
            return None

    def complete_wrap(self, func, *args, **kwargs):
        """ Readline eats exceptions raised by completer functions. """
        # Workaround readline's one-time-read of terminal width.
        termcols = shutil.get_terminal_size()[0]
        readline.parse_and_bind('set completion-display-width %d' % termcols)
        try:
            return func(*args, **kwargs)
        except:
            traceback.print_exc()
            raise

    def complete_names(self, text, line, begin, end):
        choices = set(self.root_command.subcommands)
        return [x for x in choices if x.startswith(line)]

    @contextlib.contextmanager
    def setup_readline(self):
        """ Configure our tab completion settings for a context and then
        restore them to previous settings on exit. """
        readline.parse_and_bind('tab: complete')
        completer_save = readline.get_completer()
        delims_save = readline.get_completer_delims()
        delims = set(delims_save)
        delims |= self.completer_delim_includes
        delims -= self.completer_delim_excludes
        readline.set_completer(self.completer_hook)
        try:
            readline.set_completer_delims(''.join(delims))
            try:
                yield
            finally:
                readline.set_completer_delims(delims_save)
        finally:
            readline.set_completer(completer_save)

    def run_loop(self):
        """ Main entry point for running in interactive mode. """
        self.root_command.prog = ''
        history_file = self.load_history()
        rendering.vtmlprint(self.intro)
        try:
            self.loop()
        finally:
            readline.write_history_file(history_file)

    def loop(self):
        """ Inner loop for interactive mode.  Do not call directly. """
        while True:
            with self.setup_readline():
                try:
                    line = input(self.prompt)
                except EOFError:
                    _vprinterr('^D')
                    break
                except KeyboardInterrupt:
                    _vprinterr('^C')
                    continue
            if not line.strip():
                continue
            try:
                cmd, args = self.cmd_split(line)
            except KeyError as e:
                _vprinterr('<red>Invalid command: %s</red>' % e)
                continue
            try:
                cmd(argv=args)
            except SessionExit:
                break
            except SystemExit as e:
                pass



"""
A Frankenstein combo of advanced style and simple (autocommand).
"""

import shellish


@shellish.autocommand
def subcommand1(firstarg, second, *args, optional=1, optionalstr: str=None,
                mustbeint: int=None, **kwargs: bool):
    print("Hi from sub1", firstarg, second, args, optional, optionalstr,
          mustbeint, kwargs)


@shellish.autocommand
def subcommand2(firstarg, second, *args, optional=1, optionalstr: str=None,
                mustbeint: int=None, **kwargs: bool):
    print("Hi from sub2", firstarg, second, args, optional, optionalstr,
          mustbeint, kwargs)


class Root(shellish.Command):
    """ Shellify some autocommands. """

    name = 'root'

    def setup_args(self, parser):
        self.add_subcommand(subcommand1)
        self.add_subcommand(subcommand2)

root = Root()
root()




import argparse
import io
import os
import statistics
import sys
import unittest
from shellish import layout as L


def calc_table(*columns, width=100, data=None, flex=False, **kwargs):
    t = L.Table(columns=columns, width=width, flex=flex, column_padding=0,
                **kwargs)
    renderer = t.make_renderer()
    renderer.print(data or [])
    return renderer.widths


def fileredir(call, *args, **kwargs):
    """ Override `file` with stringio object and return a tuple of a function
    to get the output and the instantiated callable. """
    file = io.StringIO()
    output = lambda: file.getvalue().splitlines()
    return output, call(*args, file=file, **kwargs)


class TableUnflex(unittest.TestCase):

    def test_only_pct(self):
        widths = calc_table(.10, .40, .50)
        self.assertEqual(widths, [10, 40, 50])

    def test_only_fixed(self):
        widths = calc_table(10, 40, 50)
        self.assertEqual(widths, [10, 40, 50])

    def test_fixed_and_pct(self):
        widths = calc_table(.10, 40, .50)
        self.assertEqual(widths, [10, 40, 50])

    def test_uneven_pct(self):
        widths = calc_table(1 / 3, 1 / 3, 1 / 3)
        self.assertEqual(widths, [33, 33, 33])

    def test_only_unspec_even(self):
        widths = calc_table(None, None, None, None)
        self.assertEqual(widths, [25, 25, 25, 25])

    def test_only_unspec_odd(self):
        widths = calc_table(None, None, None)
        self.assertEqual(widths, [33, 34, 33])

    def test_carry_over(self):
        widths = calc_table(1 / 3, 1 / 3, 1 / 3)
        self.assertEqual(widths, [33, 33, 33])

    def test_only_unspec_one(self):
        widths = calc_table(None)
        self.assertEqual(widths, [100])

    def test_only_unspec_two(self):
        widths = calc_table(None, None)
        self.assertEqual(widths, [50, 50])

    def test_mixed(self):
        widths = calc_table(25, .25, None, None)
        self.assertEqual(widths, [25, 25, 25, 25])

    def test_mixed_odd(self):
        widths = calc_table(19, 2 / 3, None, None, None, None, width=147)
        self.assertEqual(widths, [19, 98, 7, 8, 7, 8])


class TableRendering(unittest.TestCase):

    def table(self, *args, column_padding=0, **kwargs):
        self.get_lines, t = fileredir(L.Table, *args,
                                      column_padding=column_padding, **kwargs)
        return t

    def test_show_mode(self):
        t = self.table([10], width=10, clip=False, flex=False)
        text = 'A' * 20
        t.print_row([text])
        res = self.get_lines()[0]
        self.assertEqual(res, text)

    def test_clip_mode_no_cliptext(self):
        t = self.table([10], width=10, clip=True, cliptext='', flex=False)
        fits = 'A' * 10
        clipped = 'B' * 10
        t.print([[fits + clipped]])
        res = self.get_lines()[0]
        self.assertEqual(res, fits)

    def test_clip_mode_with_cliptext(self):
        for cliptext in ('*', '**', '***'):
            t = self.table([10], width=10, clip=True, cliptext=cliptext,
                           flex=False)
            fits = 'A' * (10 - len(cliptext))
            clipped = 'B' * 10
            t.print([[fits + clipped]])
            res = self.get_lines()[0]
            self.assertEqual(res, fits + cliptext)

    def test_flex_smoosh(self):
        t = self.table([None, None, None], width=12)
        text_a = ['1', '22', '333']
        text_b = ['333', '4444', '55555']
        t.print([text_a, text_b])
        first, second = self.get_lines()
        self.assertEqual(len(first.split()), 3)
        self.assertEqual(second, ''.join(text_b))
        self.assertEqual(len(second.split()), 1)
        t = self.table(column_padding=1, width=15)
        t.print([text_a, text_b])
        first, second = self.get_lines()
        self.assertEqual(second, '333 4444 55555')
        t = self.table(column_padding=2, width=18)
        t.print([text_a, text_b])
        first, second = self.get_lines()
        self.assertEqual(second, ' 333  4444  55555')
        t = self.table(column_padding=3, width=21)
        t.print([text_a, text_b])
        first, second = self.get_lines()
        self.assertEqual(second, ' 333   4444   55555')

    def test_minwidth_pct(self):
        cols = ({
            "minwidth": .10,
        }, {
            "minwidth": .40,
        }, {
            "minwidth": .50,
        })
        data = ["A" * 70, "B" * 156, "C" * 56]  # overflow all cols
        widths = calc_table(*cols, flex=True, data=data)
        self.assertEqual(widths, [10, 40, 50])

    def test_minwidth_fixed(self):
        cols = ({
            "minwidth": 10,
        }, {
            "minwidth": 40,
        }, {
            "minwidth": 50,
        })
        data = ["A" * 70, "B" * 156, "C" * 56]  # overflow all cols
        widths = calc_table(*cols, flex=True, data=data)
        self.assertEqual(widths, [10, 40, 50])


class TableDataSupport(unittest.TestCase):

    def table(self, *args, **kwargs):
        return fileredir(L.Table, *args, **kwargs)

    def tabulate(self, *args, **kwargs):
        return fileredir(L.tabulate, *args, **kwargs)

    def test_columns_from_only_list_data(self):
        output, t = self.table()
        t.print([['one', 'two', 'three']])
        output, t = self.table()
        t.print([['one', 'two', 'three']] * 10)

    def test_columns_from_only_generator_data(self):
        output, t = self.table()

        def gen():
            yield ['one', 'two', 'three']
        t.print(gen())
        self.assertEqual(len(output()), 1)

        output, t = self.table()

        def gen():
            yield ['one', 'two', 'three']
            yield ['one', 'two', 'three']
        t.print(gen())
        self.assertEqual(len(output()), 2)

    def test_columns_from_headers(self):
        output, t = self.table(headers=['One', 'Two', 'Three'])
        t.print([['one', 'two', 'three']])

    def test_columns_from_accessors(self):
        output, t = self.table(accessors=['one', 'two', 'three'])
        t.print([{'one': 'ONE', 'two': 'TWO', 'three': 'THREE'}])

    def test_columns_width_spec_only(self):
        output, t = self.table(columns=[None, None, None])
        t.make_renderer()

    def test_columns_empty_style_spec(self):
        output, t = self.table(columns=[{}, {}, {}])
        t.make_renderer()

    def test_empty_iter(self):
        output, t = self.table([None])
        t.print(iter([]))
        self.assertFalse(output())

    def test_empty_list(self):
        output, t = self.table([None])
        t.print([])
        self.assertFalse(output())

    def test_no_double_up_tabulate(self):
        output, t = self.tabulate([['abc']])
        self.assertEqual(''.join(output()).count('abc'), 1)
        output, t = self.tabulate([['abc'], ['XYZ']])
        val = ''.join(output())
        self.assertEqual(val.count('abc'), 1)
        self.assertEqual(val.count('XYZ'), 1)

    def test_dict_tabulate(self):
        output, t = self.tabulate([{
            "this_is_a_snake": "foo"
        }])
        self.assertEqual(t.headers[0], 'This Is A Snake')
        val = ''.join(output())
        self.assertIn('foo', val)
        self.assertEqual(val.count('foo'), 1)

    def test_tabulate_with_headers_empty(self):
        output, t = self.tabulate([], headers=['one'])
        self.assertIn('one', ''.join(output()))

    def test_tabulate_with_headers(self):
        output, t = self.tabulate([['ONE']], headers=['one'])
        self.assertIn('one', output()[0])
        self.assertIn(L.PlainTableRenderer.linebreak, output()[1])
        self.assertIn('ONE', output()[2])

    def test_empty_iter_tabulate(self):
        self.tabulate(iter([['header-only']]))

    def test_empty_iter_tabulate_headerarg(self):
        self.tabulate(iter([]), header=False)

    def test_empty_list_tabulate(self):
        self.tabulate([['header-only']])

    def test_empty_list_tabulate_headerarg(self):
        self.tabulate([], header=False)

    def test_empty_list_print(self):
        self.tabulate([], header=False)

    def test_generator_tabulate_headerless(self):
        def g():
            for x in range(2):
                yield [x]
        output, t = self.tabulate(g(), header=False)
        self.assertIn('1', ''.join(output()))

    def test_add_footers_no_body(self):
        out, t = self.table(headers=['One'])
        t.print_footer('foo')
        self.assertIn('One', out()[0])
        self.assertIn('foo', out()[-1])

    def test_footers_alt_renderers(self):
        for renderer in 'json', 'csv', 'plain', 'md':
            output, table = self.table(renderer=renderer)
            with table:
                table.print(['AB'])
                table.print_footer('foobar')

    def test_gen_seed_over_under(self):
        for seed_max in range(12):
            output, t = self.table([None], width=1, column_padding=0)
            t.max_seed = seed_max

            def stream():
                for i in range(10):
                    yield [i]
            t.print(stream())
            self.assertEqual([int(x) for x in output()], list(range(10)))

    def test_list_seed_over_under(self):
        for seed_max in range(12):
            output, t = self.table([None], width=1, column_padding=0)
            t.max_seed = seed_max
            data = [[i] for i in range(10)]
            t.print(data)
            self.assertEqual([int(x) for x in output()], list(range(10)))

    def test_empty_iter_noflex(self):
        output, t = self.table([None], flex=False)
        t.print(iter([]))
        self.assertFalse(output())

    def test_empty_list_noflex(self):
        output, t = self.table([20], flex=False)
        t.print([])
        self.assertFalse(output())

    def test_gen_seed_over_under_noflex(self):
        for seed_max in range(12):
            output, t = self.table([None], flex=False, width=1,
                                   column_padding=0)
            t.max_seed = seed_max

            def stream():
                for i in range(10):
                    yield [i]
            t.print(stream())
            self.assertEqual([int(x) for x in output()], list(range(10)))

    def test_list_seed_over_under_noflex(self):
        for seed_max in range(12):
            output, t = self.table([1], flex=False, width=1, column_padding=0)
            t.max_seed = seed_max
            data = [[i] for i in range(10)]
            t.print(data)
            self.assertEqual([int(x) for x in output()], list(range(10)))


class TableUsagePatterns(unittest.TestCase):

    def table(self, *args, column_padding=0, **kwargs):
        return fileredir(L.Table, *args, column_padding=column_padding,
                         **kwargs)

    def test_headers_once(self):
        output, t = self.table(headers=['foo'], width=3)
        t.print([['one']])
        t.print([['two']])
        self.assertEqual(output(), ['foo', ('\u2014' * 3), 'one', 'two'])


class TableCalcs(unittest.TestCase):

    def test_unflex_spec_underflow(self):
        widths = calc_table(*[1 / 26] * 26)
        self.assertLess(statistics.variance(widths), 1)
        self.assertEqual(sum(widths), 78)  # uses floor() so it's lossy

    def test_unflex_unspec_underflow(self):
        widths = calc_table(*[None] * 26)
        self.assertLess(statistics.variance(widths), 1)
        self.assertEqual(sum(widths), 100)

    def test_equal_flex_underflow_all_fits(self):
        widths = calc_table(*[None] * 26, flex=True, data=[['a'] * 26])
        self.assertLess(statistics.variance(widths), 1)
        self.assertEqual(sum(widths), 100)

    def test_uniform_dist(self):
        dist = L.VisualTableRenderer._uniform_dist
        for i in range(1, 101):
            for ii in range(151):
                d = dist(None, i, ii)
                self.assertEqual(sum(d), ii, (i, ii, d))

    def test_multiline_flex(self):
        data = [('aaa\naaaaaa', 'aaaaaa')]
        widths = calc_table(None, None, flex=True, data=data, width=12)
        self.assertListEqual(widths, [6, 6])

    def test_multiline_preformatted(self):
        data = [('aaa\naaaaaa', 'aaaaaa')]
        widths = calc_table({"overflow": "preformatted"}, {"overflow": "wrap"},
                            flex=True, data=data, width=10)
        self.assertListEqual(widths, [6, 4])
        widths = calc_table({"overflow": "preformatted"}, {"overflow": "clip"},
                            flex=True, data=data, width=10)
        self.assertListEqual(widths, [6, 4])
        widths = calc_table({"overflow": "preformatted"}, {"overflow": "wrap"},
                            flex=True, data=data, width=6)
        self.assertListEqual(widths, [6, 1])


class TableArgGroup(unittest.TestCase):

    def test_table_group(self):
        L.Table.attach_arguments(argparse.ArgumentParser())


class TableClosingContext(unittest.TestCase):

    def test_Table_context_noaction(self):
        closed = False

        class TestTable(L.Table):
            def close(self, **kwargs):
                super().close(*kwargs)
                nonlocal closed
                closed = True
        with TestTable():
            pass
        self.assertTrue(closed)

    def test_Table_context_exc(self):
        class TestExc(Exception):
            pass

        class TestTable(L.Table):
            def close(this, exception=None):
                self.assertIs(exception[0], TestExc)
                super().close(exception)
        try:
            with TestTable():
                raise TestExc()
        except TestExc:
            pass


class JSONTableRenderer(unittest.TestCase):

    def setUp(self):
        t = L.Table([None], renderer='json')
        self.r = t.make_renderer()

    def test_make_key_snakecase(self):
        self.assertEqual(self.r.make_key('snake_case'), 'snakeCase')
        self.assertEqual(self.r.make_key('snake_case_more'), 'snakeCaseMore')

    def test_make_key_snakecase_dblunderscore(self):
        self.assertEqual(self.r.make_key('snake__case'), 'snakeCase')

    def test_make_key_snakecase_leadunderscore(self):
        self.assertEqual(self.r.make_key('_snake_case'), 'SnakeCase')

    def test_make_key_snakecase_dblleadunderscore(self):
        self.assertEqual(self.r.make_key('__snake_case'), 'SnakeCase')

    def test_make_key_dupkeys(self):
        self.assertEqual(self.r.make_key('foo'), 'foo')
        self.assertEqual(self.r.make_key('foo'), 'foo1')
        self.assertEqual(self.r.make_key('foo'), 'foo2')
        self.assertEqual(self.r.make_key('foo2'), 'foo21')

    def test_make_key_scrub(self):
        self.assertEqual(self.r.make_key('!@#$%^&*()foo'), 'foo')
        self.assertEqual(self.r.make_key('!@#$%^&*()foo:"bar'), 'foobar')

    def test_make_key_space(self):
        self.assertEqual(self.r.make_key('This is some space - with-dashes'),
                         'thisIsSomeSpaceWithDashes')


class TreeData(unittest.TestCase):
    """ Tests to make sure tree doesn't blow up with various data inputs. """

    def setUp(self):
        self.stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def tearDown(self):
        sys.stdout.close()
        sys.stdout = self.stdout

    def test_treeprint_empty(self):
        self.assertRaises(TypeError, L.treeprint)

    def test_treeprint_dict(self):
        L.treeprint({})
        L.treeprint({1: 1})
        L.treeprint({"1": 1})
        L.treeprint({"1": "1"})
        L.treeprint({1: "1"})
        L.treeprint({"a": "1"})
        L.treeprint({"a": 1})
        L.treeprint({"a": [1, 2]})
        L.treeprint({"a": ["1", 2]})
        L.treeprint({"a": ["1", "a"]})
        L.treeprint({"a": {}})
        L.treeprint({"a": {1: 1}})

    def test_treeprint_cyclic(self):
        looper = {}
        looper['myself'] = looper
        self.assertRaises(ValueError, L.treeprint, looper)

    def test_treeprint_list(self):
        L.treeprint([])
        L.treeprint([1, 2])
        L.treeprint(["a", "b"])
        L.treeprint([{}, "b"])
        L.treeprint([{1, 2}, "b"])
        L.treeprint([[]])
        L.treeprint([[1]])
        L.treeprint([1, [1]])
        L.treeprint([None])

    def test_treeprint_tuple(self):
        L.treeprint(())
        L.treeprint((1, 2))
        L.treeprint(("a", "b"))
        L.treeprint(({}, {}))

    def test_treeprint_numbers(self):
        L.treeprint(1)
        L.treeprint(0)

    def test_treeprint_odd(self):
        L.treeprint(None)
        L.treeprint(True)
        L.treeprint(False)
        L.treeprint(object())
        L.treeprint(object)



"""
Supplemental code for stdlib package(s).  Namely argparse.
"""

import argparse
import io
import os
import re
import shutil
import sys
import warnings
from .. import paging, layout
from ..rendering import vtmlrender, VTMLBuffer


class HelpSentinel(str):

    def __len__(self):
        return 1

HELP_SENTINEL = HelpSentinel()
NBSP = ' '  # U+00A0


class ShellishHelpFormatter(argparse.HelpFormatter):

    leadingws = re.compile('^\s+')
    whitespace = re.compile('[ \n\t\v\f\r]+')
    max_width = 100

    class _Section(argparse.HelpFormatter._Section):

        def format_help(self):
            if self.parent is not None:
                self.formatter._indent()
            join = self.formatter._join_parts
            item_help = join([func(*args) for func, args in self.items])
            if self.parent is not None:
                self.formatter._dedent()
            if not item_help:
                return ''
            if self.heading in (argparse.SUPPRESS, None):
                heading_bar = ''
                return item_help
            else:
                current_indent = self.formatter._current_indent
                heading = '%*s%s' % (current_indent, '', self.heading.upper())
                heading_bar = vtmlrender('\n<u>%s</u>:\n\n' % heading)
            return join([heading_bar, item_help])

    def __init__(self, prog, max_width=None, width=None, **kwargs):
        if width is None:
            if max_width is None:
                max_width = self.max_width
            width = min(shutil.get_terminal_size()[0] - 2, max_width)
        super().__init__(prog, width=width, **kwargs)

    def format_help(self):
        return self._root_section.format_help()

    def _fill_text(self, text, width=None, indent=None):
        """ Reflow text width while maintaining certain formatting
        characteristics like double newlines and indented statements. """
        assert isinstance(text, str)
        if indent is None:
            indent = NBSP * self._current_indent
        assert isinstance(indent, str)
        paragraphs = []
        line_buf = []
        pre = ''
        for fragment in text.splitlines():
            pre_indent = self.leadingws.match(fragment)
            if not fragment or pre_indent:
                if line_buf:
                    line = ' '.join(line_buf)
                    paragraphs.append((pre, self.whitespace.sub(' ', line)))
                if not fragment:
                    paragraphs.append(('', ''))
                else:
                    pre = pre_indent.group()
                    fragment = self.leadingws.sub('', fragment)
                    paragraphs.append((pre, fragment))
                line_buf = []
                pre = ''
            else:
                line_buf.append(fragment)
        if line_buf:
            line = ' '.join(line_buf)
            paragraphs.append((pre, self.whitespace.sub(' ', line)))
        indent = VTMLBuffer(indent)
        nl = VTMLBuffer('\n')
        if width is None:
            width = self._width - len(indent)
        lines = []
        for pre, paragraph in paragraphs:
            pwidth = width - len(pre)
            lines.append(nl.join((indent + pre + x)
                         for x in vtmlrender(paragraph).wrap(pwidth)))
        return nl.join(lines)

    def _format_text(self, text):
        if '%(prog)' in text:
            text = text % dict(prog=self._prog)
        return self._fill_text(text) + '\n\n'

    def _get_help_string(self, action):
        """ Adopted from ArgumentDefaultsHelpFormatter. """
        raise NotImplementedError('')

    def _format_usage(self, *args, **kwargs):
        usage = super()._format_usage(*args, **kwargs)
        return VTMLBuffer('\n').join(vtmlrender('<red>%s</red>' % x)
                                     for x in usage.split('\n'))

    def _join_parts(self, parts):
        return VTMLBuffer('').join(x for x in parts
                                   if x not in (None, argparse.SUPPRESS))

    def _get_type_label(self, typeattr):
        if isinstance(typeattr, argparse.FileType) or typeattr is open:
            return 'file'
        if hasattr(typeattr, '__name__'):
            name = typeattr.__name__
            if name != '<lambda>':
                return name

    def _get_default_metavar_for_optional(self, action):
        if action.type is None:
            return 'STR'
        label = self._get_type_label(action.type)
        if label is not None:
            return label.upper()
        return 'VALUE'

    def _get_default_metavar_for_positional(self, action):
        if action.dest is argparse.SUPPRESS:
            if isinstance(action, argparse._SubParsersAction):
                return action.metavar or 'SUBCOMMAND'
            else:
                raise RuntimeError("Can not produce metavar for: %r" % action)
        return action.dest.upper()

    def _format_action_invocation(self, action):
        args = None
        if not action.option_strings:  # positional
            if action.metavar is None:
                options = [self._get_default_metavar_for_positional(action)]
            elif action.metavar is not argparse.SUPPRESS:
                options = [action.metavar]
            else:
                raise NotImplementedError()
        else:
            options = action.option_strings
            if action.nargs != 0:
                metavar = self._get_default_metavar_for_optional(action)
                args = self._format_args(action, metavar)
        res = ', '.join('<b><blue>%s</blue></b>' % x for x in options)
        return ' '.join((res, args)) if args else res

    def _metavar_formatter(self, action, default_metavar):
        if action.metavar is not None:
            result = action.metavar
        else:
            result = default_metavar

        def format(tuple_size):
            if isinstance(result, tuple):
                return result
            else:
                return (result, ) * tuple_size
        return format

    def _format_action_table(self, action, terse=False):
        indent = NBSP * self._current_indent
        pad = NBSP * 2
        is_subparser = isinstance(action, argparse._SubParsersAction)
        if is_subparser:
            sub_name = self._get_default_metavar_for_positional(action)
            left_col = ["<b>%s</b>" % sub_name]
        else:
            left_col = [self._format_action_invocation(action)]
        if action.required:
            left_col[-1] += ' <b><i>(required)</i></b>'
        if action.default not in (argparse.SUPPRESS, None) and \
           action.nargs != 0:
            if isinstance(action.default, io.IOBase):
                default = action.default.name
            else:
                default = action.default
            left_col.append(pad + 'default: <b>%s</b>' % (default,))
        if not terse and not is_subparser:
            if isinstance(action.nargs, int):
                arg_count = action.nargs
            else:
                arg_count = {
                    argparse.OPTIONAL: 'optional',
                    argparse.ZERO_OR_MORE: 'zero or more',
                    argparse.ONE_OR_MORE: 'one or more',
                    argparse.REMAINDER: 'remainder',
                }.get(action.nargs)
            if arg_count:
                left_col.append(pad + 'count: <b>%s</b>' % arg_count)
            label = self._get_type_label(action.type)
            if label:
                left_col.append(pad + 'type: <b>%s</b>' % label)
            if action.choices:
                if len(action.choices) < 5:
                    choices = ', '.join('<b>%s</b>' % x
                                        for x in action.choices)
                    left_col.append(pad + 'choices: {%s}' % choices)
                else:
                    left_col.append(pad + 'choices:')
                    for x in action.choices:
                        left_col.append((pad * 3) + '<b>%s</b>' % x)
            if getattr(action, 'env', None):
                left_col.append(pad + 'env: <b>%s</b>' % action.env)
        left_col = [indent + x for x in left_col]
        # Perform optional substitutions on help text.
        if action.help and action.help is not HELP_SENTINEL:
            params = dict(vars(action), prog=self._prog)
            for name in list(params):
                if params[name] is argparse.SUPPRESS:
                    del params[name]
            for name in list(params):
                if hasattr(params[name], '__name__'):
                    params[name] = params[name].__name__
            if params.get('choices') is not None:
                choices_str = ', '.join([str(c) for c in params['choices']])
                params['choices'] = choices_str
            help_text = action.help % params
        else:
            help_text = ''
        feed = [('\n'.join(left_col), help_text)]
        for subaction in self._iter_indented_subactions(action):
            feed.extend(self._format_action_table(subaction, terse=True))
        return feed

    def add_table(self, actions):
        table_output = io.StringIO()
        table_output.isatty = sys.stdout.isatty
        config = {
            "width": self._width,
            "justify": True,
            "overflow": 'wrap',
            "columns": [{
                "padding": 2,
                "overflow": 'preformatted'
            }, {
                "padding": 4
            }]
        }
        table = layout.Table(file=table_output, **config)
        rows = []
        for action in actions:
            if action.help is argparse.SUPPRESS:
                continue
            rows.extend(self._format_action_table(action))

        def render(data):
            table.print(data)
            table.close()
            return table_output.getvalue()
        if rows:
            self._add_item(render, [rows])


class ShellishParser(argparse.ArgumentParser):

    HelpFormatter = ShellishHelpFormatter

    def __init__(self, *args, command=None, formatter_class=None, **kwargs):
        self._env_actions = {}
        self._command = command
        if formatter_class is None:
            formatter_class = self.HelpFormatter
        super().__init__(*args, formatter_class=formatter_class, **kwargs)

    def bind_env(self, action, env):
        """ Bind an environment variable to an argument action.  The env
        value will traditionally be something uppercase like `MYAPP_FOO_ARG`.

        Note that the ENV value is assigned using `set_defaults()` and as such
        it will be overridden if the argument is set via `parse_args()` """
        if env in self._env_actions:
            raise ValueError('Duplicate ENV variable: %s' % env)
        self._env_actions[env] = action
        action.env = env

    def unbind_env(self, action):
        """ Unbind an environment variable from an argument action.  Only used
        when the subcommand hierarchy changes. """
        del self._env_actions[action.env]
        delattr(action, 'env')

    def parse_known_args(self, *args, **kwargs):
        env_defaults = {}
        for env, action in self._env_actions.items():
            if os.environ.get(env):
                env_defaults[action.dest] = os.environ[env]
                action.required = False  # XXX This is a hack
        if env_defaults:
            self.set_defaults(**env_defaults)
        return super().parse_known_args(*args, **kwargs)

    def format_help(self):
        formatter = self._get_formatter()
        formatter.add_usage(self.usage, self._actions,
                            self._mutually_exclusive_groups)
        if self.description and '\n' in self.description:
            desc = self.description.split('\n\n', 1)
            if len(desc) == 2 and '\n' not in desc[0]:
                title, about = desc
            else:
                title, about = '', desc
        else:
            title, about = self.description, ''
        if title and title.strip():
            formatter.add_text('<b><u>%s</u></b>\n' % title.strip())
        if about and about.rstrip():
            formatter.add_text(about.rstrip())
        for group in self._action_groups:
            formatter.start_section(group.title)
            formatter.add_text(group.description)
            formatter.add_table(actions=group._group_actions)
            formatter.end_section()
        formatter.add_text(self.epilog)
        return formatter.format_help()

    def print_help(self, *args, **kwargs):
        """ Add pager support to help output. """
        if self._command is not None and self._command.session.allow_pager:
            desc = 'Help\: %s' % '-'.join(self.prog.split())
            pager_kwargs = self._command.get_pager_spec()
            with paging.pager_redirect(desc, **pager_kwargs):
                return super().print_help(*args, **kwargs)
        else:
                return super().print_help(*args, **kwargs)

    def _print_message(self, message, file=None):
        if not message:
            return
        if file is None:
            file = sys.stderr
        if isinstance(message, VTMLBuffer) and not file.isatty():
            message = message.plain()
        file.write(str(message))

    def add_argument(self, *args, help=HELP_SENTINEL, **kwargs):
        return super().add_argument(*args, help=help, **kwargs)

    def add_subparsers(self, prog=None, **kwargs):
        """ Supplement a proper `prog` keyword argument for the subprocessor.
        The superclass technique for getting the `prog` value breaks because
        of our VT100 escape codes injected by `format_help`. """
        if prog is None:
            # Use a non-shellish help formatter to avoid vt100 codes.
            f = argparse.HelpFormatter(prog=self.prog)
            f.add_usage(self.usage, self._get_positional_actions(),
                        self._mutually_exclusive_groups, '')
            prog = f.format_help().strip()
        return super().add_subparsers(prog=prog, **kwargs)

    def _get_positional_kwargs(self, dest, **kwargs):
        dest = dest.replace('-', '_')
        return super()._get_positional_kwargs(dest, **kwargs)


class SafeFileContext(object):
    """ Used by SafeFileType to provide a file-like context manager. """

    def __init__(self, ft, filename):
        self.ft = ft
        self.filename = filename
        self.fd = None
        self.is_stdio = None
        self.used = False

    def __call__(self):
        warnings.warn("Calling the file argument is no longer required")
        return self

    def __enter__(self):
        assert not self.used
        self.used = True
        if self.filename == '-':
            self.is_stdio = True
            if 'r' in self.ft._mode:
                stdio = sys.stdin
            elif 'w' in self.ft._mode:
                stdio = sys.stdout
            else:
                raise ValueError("Invalid mode for stdio: %s" % self.ft._mode)
            self.fd = stdio
        else:
            self.is_stdio = False
            self.fd = open(self.filename, self.ft._mode, self.ft._bufsize,
                           self.ft._encoding, self.ft._errors)
        return self.fd

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.fd is not None:
            if self.is_stdio:
                self.fd.flush()
            else:
                self.fd.close()

    def __str__(self):
        """ Report the last string passed into our call.  This is our candidate
        filename but in practice it is THE filename used. """
        return str(self.filename)

    def __repr__(self):
        """ Report the last string passed into our call.  This is our candidate
        filename but in practice it is THE filename used. """
        return '<%s: %s>' % (type(self).__name__, repr(self.filename))


class SafeFileType(argparse.FileType):
    """ A side-effect free version of argparse.FileType that prevents erroneous
    creation of files when doing tab completion.  Arguments that use this type
    are given a factory function that will return a context manager for the
    underlying file. """

    def __call__(self, string):
        return SafeFileContext(self, string)



"""
Decorate a function that uses keyword arguments.
"""

import shellish


@shellish.autocommand
def hello(firstarg, second, *args, optional=1, optionalstr: str=None,
          mustbeint: int=None, **kwargs: bool):
    print("Hello", firstarg, second, args, optional, optionalstr,
          mustbeint, kwargs)

hello()



"""
Demo how to add your auto tab completers.
"""

import random
import shellish


def thing_completer(prefix, args):
    letters = 'qwertyuiopasdfghjklzxcvbnm'
    word = lambda: ''.join(random.sample(letters, random.randint(1, 16)))
    return set(word() for x in range(random.randint(1, 1000)))

thing = shellish.Command(name='thing', title='Demo Tab Completion')
thing.add_argument('--choices', choices=['one', 'two', 'three'])
thing.add_argument('--function', complete=thing_completer)
root = shellish.Command(name='root')
root.add_subcommand(thing)
root.get_or_create_session().run_loop()



import shellish


class Hello(shellish.Command):
    """ I am a docstring used to document the --help output! """

    name = 'hello'


class World(shellish.Command):
    """ Say something. """

    name = 'world'

    def run(self, args):
        print('Hello World')


hello = Hello()
hello.add_subcommand(World)
hello()
